#
# Qt gui for viewing DataBlock2d data files.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections
import configparser
import os
import sys
import traceback

import numpy as np

import myokit
import myokit.gui

from myokit.gui import QtWidgets, QtGui, QtCore, Qt


# Application title
TITLE = 'Myokit DataBlock Viewer'

# Settings file
SETTINGS_FILE = os.path.join(myokit.DIR_USER, 'DataBlockViewer.ini')

# Number of recent files to display
N_RECENT_FILES = 5

# About
ABOUT = f'<h1>{TITLE}</h1>' + f"""
<p>
    Myokit's DataBlock viewer is a utility to examine
    <code>DataBlock1d</code> and <code>DataBlock2d</code> objects
    containing simulation results for rectangular grids of cells.
</p>
<p>
    DataBlocks can be saved to disk and then loaded into the viewer.
    Tissue-response to different 2d variables can be animated, and
    individual cells traces can be viewed in the graph area below. To
    select a cell, click anywhere in the animation.
</p>
<p>
    System info:
    <br />Python: {sys.version}
    <br />Using the {myokit.gui.backend} GUI backend.
</p>
"""

# License
LICENSE = myokit.LICENSE_HTML

# File filters
FILTER_CSV = 'Comma separated file (*.txt *.csv)'
FILTER_IMG = ';;'.join([
    'Type detected by extension (*.png *.jpg *.jpeg *.bmp)',
    'BMP (*.bmp)',
    'JPEG (*.jpeg *.jpg)',
    'PNG (*.png)',
])
IMAGE_TYPES = ['PNG', 'JPG', 'JPEG', 'BMP']


# Application icon
def icon():
    icons = [
        'icon-datablock-viewer.ico',
        'icon-datablock-viewer-16.xpm',
        'icon-datablock-viewer-24.xpm',
        'icon-datablock-viewer-32.xpm',
        'icon-datablock-viewer-48.xpm',
        'icon-datablock-viewer-64.xpm',
        'icon-datablock-viewer-96.xpm',
        'icon-datablock-viewer-128.xpm',
        'icon-datablock-viewer-256.xpm',
    ]
    icon = QtGui.QIcon()
    for i in icons:
        icon.addFile(os.path.join(myokit.DIR_DATA, 'gui', i))
    return icon


class DataBlockViewer(myokit.gui.MyokitApplication):
    """
    Graphical interface for viewing DataBlock data.
    """

    def __init__(self, filename=None):
        super().__init__()
        # Set application icon
        self.setWindowIcon(icon())

        # Set size, center
        self.resize(800, 600)
        self.setMinimumSize(600, 400)
        qr = self.frameGeometry()
        cp = QtGui.QGuiApplication.primaryScreen().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

        # Status bar
        self._label_cursor = QtWidgets.QLabel()
        self.statusBar().addPermanentWidget(self._label_cursor)
        self.statusBar().showMessage('Ready')

        # Menu bar
        self.create_menu()

        # Timer
        self._timer_interval = 50
        self._timer = QtCore.QTimer(self)
        self._timer.setInterval(self._timer_interval)
        self._timer.setSingleShot(False)
        self._timer.timeout.connect(self.action_next_frame)
        self._timer_paused = False
        self._timer.setTimerType(Qt.TimerType.PreciseTimer)

        # Video widget
        self._video_scene = VideoScene()
        self._video_scene.mouse_moved.connect(self.event_video_mouse_move)
        self._video_scene.single_click.connect(self.event_video_single_click)
        self._video_scene.double_click.connect(self.event_video_double_click)
        self._video_view = VideoView(self._video_scene)
        self._video_view.setMouseTracking(True)
        self._video_view.setCursor(Qt.CursorShape.CrossCursor)
        #self._video_view.setViewport(QtOpenGL.QGLWidget())
        self._video_pixmap = None
        self._video_item = None
        self._video_frames = None
        self._video_iframe = None

        # Colorbar widget
        self._colorbar_width = 20
        self._colorbar_height = 256
        self._colorbar_scene = VideoScene()
        self._colorbar_view = VideoView(self._colorbar_scene)
        self._colorbar_view.setMaximumWidth(self._colorbar_width)
        self._colorbar_pixmap = None
        self._colorbar_item = None

        # Video slider
        self._slider = QtWidgets.QSlider(Qt.Orientation.Horizontal)
        self._slider.setTickPosition(QtWidgets.QSlider.TickPosition.NoTicks)
        self._slider.setSingleStep(1)
        self._slider.setMinimum(0)
        self._slider.setMaximum(0)
        self._slider.sliderPressed.connect(self.action_pause_timer)
        self._slider.sliderReleased.connect(self.action_depause_timer)
        self._slider.valueChanged.connect(self.action_set_frame)

        # Controls
        style = QtWidgets.QApplication.style()

        # Play button
        self._play_icon_play = style.standardIcon(
            style.StandardPixmap.SP_MediaPlay)
        self._play_icon_pause = style.standardIcon(
            style.StandardPixmap.SP_MediaPause)
        self._play_button = QtWidgets.QPushButton()
        self._play_button.setIcon(self._play_icon_play)
        self._play_button.pressed.connect(self.action_start_stop)

        # Frame indicator
        self._frame_label = QtWidgets.QLabel('Frame')
        self._frame_field = QtWidgets.QLineEdit('0')
        self._frame_field.setReadOnly(True)
        self._frame_field.setMaxLength(6)
        self._frame_field.setMaximumWidth(100)
        self._frame_label.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)

        # Time indicator
        self._time_label = QtWidgets.QLabel('Time')
        self._time_field = QtWidgets.QLineEdit('0')
        self._time_field.setReadOnly(True)
        self._time_field.setMaxLength(6)
        self._time_field.setMaximumWidth(100)
        self._time_label.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)

        # Speed indicator
        self._rate_label = QtWidgets.QLabel('Delay')
        self._rate_field = QtWidgets.QLineEdit(str(self._timer_interval))
        self._rate_field.setValidator(QtGui.QIntValidator(1, 2**20, self))
        self._rate_field.editingFinished.connect(self.event_rate_changed)
        self._rate_field.setMaximumWidth(100)
        self._rate_label.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)

        # Graph controls
        self._graph_clear_button = QtWidgets.QPushButton('Clear graphs')
        self._graph_clear_button.pressed.connect(self.action_clear_graphs)

        # Variable selection
        self._variable_select = QtWidgets.QComboBox()
        self._variable_select.activated.connect(self.event_variable_selected)
        self._variable_select.setMinimumWidth(160)

        # Colormap selection
        self._colormap = next(iter(myokit.ColorMap.names()))
        self._colormap_select = QtWidgets.QComboBox()
        for cmap in myokit.ColorMap.names():
            self._colormap_select.addItem(cmap)
        self._colormap_select.activated.connect(self.event_colormap_selected)
        self._colormap_select.setMinimumWidth(120)

        self._colormap_lower_label = QtWidgets.QLabel('Range')
        self._colormap_lower_label.setMaximumWidth(50)
        self._colormap_lower_label.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self._colormap_lower_field = AutoFloatField()
        self._colormap_lower_field.setMaxLength(6)
        self._colormap_lower_field.setMaximumWidth(80)
        self._colormap_lower_field.editingFinished.connect(
            self.event_variable_selected)

        self._colormap_upper_label = QtWidgets.QLabel('to')
        self._colormap_upper_label.setMaximumWidth(20)
        self._colormap_upper_label.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter)
        self._colormap_upper_field = AutoFloatField()
        self._colormap_upper_field.setMaxLength(6)
        self._colormap_upper_field.setMaximumWidth(80)
        self._colormap_upper_field.editingFinished.connect(
            self.event_variable_selected)

        # Control layout
        self._control_layout = QtWidgets.QHBoxLayout()
        self._control_layout.addWidget(self._play_button)
        self._control_layout.addWidget(self._frame_label)
        self._control_layout.addWidget(self._frame_field)
        self._control_layout.addWidget(self._time_label)
        self._control_layout.addWidget(self._time_field)
        self._control_layout.addWidget(self._rate_label)
        self._control_layout.addWidget(self._rate_field)
        self._control_layout.addWidget(self._graph_clear_button)
        self._control_layout.addWidget(self._variable_select)
        self._control_layout.addWidget(self._colormap_select)
        self._control_layout.addWidget(self._colormap_lower_label)
        self._control_layout.addWidget(self._colormap_lower_field)
        self._control_layout.addWidget(self._colormap_upper_label)
        self._control_layout.addWidget(self._colormap_upper_field)

        # Graph area
        self._graph_area = GraphArea()
        self._graph_area.mouse_moved.connect(self.event_graph_mouse_move)
        self._graph_area.setMouseTracking(True)
        self._graph_area.setCursor(Qt.CursorShape.CrossCursor)

        # Video and colorbar layout
        self._video_plus_layout = QtWidgets.QHBoxLayout()
        self._video_plus_layout.addWidget(self._video_view)
        self._video_plus_layout.addWidget(self._colorbar_view)

        # Video Layout
        self._video_layout = QtWidgets.QVBoxLayout()
        self._video_layout.addLayout(self._video_plus_layout)
        self._video_layout.addWidget(self._slider)
        self._video_layout.addLayout(self._control_layout)
        self._video_widget = QtWidgets.QWidget()
        self._video_widget.setLayout(self._video_layout)

        # Central layout
        self._central_widget = QtWidgets.QSplitter(Qt.Orientation.Vertical)
        self._central_widget.addWidget(self._video_widget)
        self._central_widget.addWidget(self._graph_area)
        self.setCentralWidget(self._central_widget)

        # Current path, current file, recent files
        self._path = QtCore.QDir.currentPath()
        self._file = None
        self._recent_files = []

        # Current data block, display variable
        self._data = None
        self._variable = None

        # Load settings from ini file
        self.load_config()
        self.update_window_title()

        # Set controls to correct values
        self._colormap_select.setCurrentIndex(
            self._colormap_select.findText(self._colormap))
        self._rate_field.setText(str(self._timer_interval))
        self._timer.setInterval(self._timer_interval)

        # Pause video playback during resize
        self._resize_timer = QtCore.QTimer()
        self._resize_timer.timeout.connect(self._resize_timeout)
        self._video_view.resize_event.connect(self._resize_started)
        self._colorbar_view.resize_event.connect(self._resize_started)

        # Attempt to load selected file
        if filename and os.path.isfile(filename):
            self.load_data_file(filename)

        # Focus on play button
        self._play_button.setFocus()

    def action_about(self):
        """
        Displays the about dialog.
        """
        QtWidgets.QMessageBox.about(self, TITLE, ABOUT)

    def action_clear_graphs(self):
        """
        Removes all graphs from the graph area.
        """
        self._graph_area.clear()

    def action_depause_timer(self):
        """
        De-pauses the timer, restarting it if was paused, not affecting it if
        it wasn't.
        """
        if self._timer_paused:
            self._timer_paused = False
            if self._data is not None:
                self._timer.start()

    def action_extract_colormap_image(self):
        """ Extracts the current colormap to an image file. """

        if self._data is None:
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>No data to export.</h1>'
                '<p>Please open a data file first.</p>')
            return

        fname = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Extract colormap to image file',
            self._path,
            filter=FILTER_IMG)[0]

        if fname:
            fname = str(fname)
            ext = os.path.splitext(fname)[1][1:].upper()
            if ext not in IMAGE_TYPES:
                QtWidgets.QMessageBox.warning(
                    self, TITLE,
                    '<h1>Image type not recognized.</h1>'
                    f'<p>Unknown image type "{ext}".</p>')
                return

            # Create image
            nx = 200
            ny = 800
            image = myokit.ColorMap.image(self._colormap, nx, ny)
            image = QtGui.QImage(
                image, nx, ny, QtGui.QImage.Format.Format_ARGB32)

            # Add lower and upper bounds
            lower = self._colormap_lower_field.value()
            upper = self._colormap_upper_field.value()
            if lower is None or upper is None:
                data = self._data.get2d(self._variable)
                lower = np.min(data) if lower is None else lower
                upper = np.max(data) if upper is None else upper
            painter = QtGui.QPainter(image)
            painter.drawText(10, 15, str(upper))
            painter.drawText(10, ny - 5, str(lower))
            painter.end()

            # Save
            image.save(fname, ext)

    def action_extract_frame(self):
        """
        Extracts the current frame to a csv file.
        """
        if self._data is None:
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>No data to export.</h1>'
                '<p>Please open a data file first.</p>')
            return
        fname = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Extract frame to csv file',
            self._path,
            filter=FILTER_CSV)[0]
        if fname:
            fname = str(fname)
            self._data.save_frame_csv(
                fname, self._variable, self._video_iframe)

    def action_extract_frame_image(self):
        """
        Extracts the current frame to an image file.
        """
        if self._data is None:
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>No data to export.</h1>'
                '<p>Please open a data file first.</p>')
            return
        fname = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Extract frame to image file',
            self._path,
            filter=FILTER_IMG)[0]
        if fname:
            fname = str(fname)
            ext = os.path.splitext(fname)[1][1:].upper()
            if ext not in IMAGE_TYPES:
                QtWidgets.QMessageBox.warning(
                    self, TITLE,
                    '<h1>Image type not recognized.</h1>'
                    f'<p>Unknown image type "{ext}".</p>')
                return
            nt, ny, nx = self._data.shape()
            image = self._video_frames[self._video_iframe]
            image = QtGui.QImage(
                image, nx, ny, QtGui.QImage.Format.Format_ARGB32)
            image.save(fname, ext)

    def action_extract_graphs(self):
        """
        Extracts the currently displayed graphs to a csv file.
        """
        if self._data is None:
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>No data to export.</h1>'
                '<p>Please open a data file first.</p>')
            return
        d = self._graph_area.log()
        if len(d) == 0:
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>No graphs to export.</h1>'
                '<p>Please add some graphs to the graph panel by clicking or'
                ' double clicking the video area.</p>')
            return
        fname = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Extract graphs to csv file',
            self._path,
            filter=FILTER_CSV)[0]
        if fname:
            fname = str(fname)
            d.save_csv(fname)

    def action_license(self):
        """
        Displays this program's licensing information.
        """
        QtWidgets.QMessageBox.about(self, TITLE, LICENSE)

    def action_next_frame(self, e=None):
        """
        Move to next frame!
        """
        self.action_set_frame(self._video_iframe + 1)

    def action_open(self):
        """
        Let the user select a data file to load.
        """
        self.action_stop_timer()
        fname = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Open data file', directory=self._path)[0]
        QtWidgets.QApplication.processEvents()
        if fname:
            fname = str(fname)
            self.load_data_file(fname)

    def action_pause_timer(self):
        """
        Pauses the timer (if running). Only meant to be used for very short
        periods!
        """
        if self._timer.isActive():
            self._timer_paused = True
        self._timer.stop()

    def action_recent_file(self):
        """
        Opens a recent file.
        """
        action = self.sender()
        if action:
            fname = str(action.data())
            if not os.path.isfile(fname):
                QtWidgets.QMessageBox.warning(
                    self, TITLE,
                    '<h1>Failed to load file.</h1>'
                    '<p>The selected file no longer exists.</p>')
            else:
                self.load_data_file(fname)

    def action_start_stop(self):
        """
        Toggles the started/stopped state of the timer.
        """
        if self._timer.isActive():
            self.action_stop_timer()
        else:
            self.action_start_timer()

    def action_set_colormap(self, name):
        """
        Loads the ColorMap specified by ``name``.

        If data is loaded, this will also call :meth:`action_set_variable` to
        update the video frames.
        """
        name = str(name)
        if not myokit.ColorMap.exists(name):
            return  # Silent return?

        # Set colormap
        self._colormap = name

        # Update colormap controls
        self._colormap_select.setCurrentIndex(
            self._colormap_select.findText(self._colormap))

        # Update colorbar
        if self._data is not None:
            nx = self._colorbar_width
            ny = self._colorbar_height
            image = myokit.ColorMap.image(self._colormap, nx, ny)
            image = QtGui.QImage(
                image, nx, ny, QtGui.QImage.Format.Format_ARGB32)
            self._colorbar_pixmap.convertFromImage(image)
            self._colorbar_item.setPixmap(self._colorbar_pixmap)  # qt5
            self.action_set_variable(self._variable)

    def action_set_variable(self, var):
        """
        Loads the variable specified by the name ``var`` into the main display.

        This method is also responsible for converting the data into frames to
        be displayed on the video widget.
        """
        if self._data is None:
            return
        self._variable = str(var)
        self._variable_select.setCurrentIndex(
            self._variable_select.findText(self._variable))
        self.action_pause_timer()
        self._video_frames = self._data.images(
            self._variable,
            self._colormap,
            self._colormap_lower_field.value(),
            self._colormap_upper_field.value()
        )
        self.action_set_frame(self._video_iframe)
        self.action_depause_timer()

    def action_set_frame(self, frame):
        """
        Move to a specific frame.

        This method updates the display to show a video frame created by an
        earlier call to :meth:`action_set_variable`.
        """
        # Check frame index
        nt, ny, nx = self._data.shape()
        frame = int(frame)
        if frame < 0 or frame >= nt:
            frame = 0
        # Update
        if self._video_iframe != frame or True:
            self._video_iframe = frame
            # Update slider
            if self._slider.value() != frame:
                self._slider.setValue(self._video_iframe)
            # Update frame/time information
            self._frame_field.setText(str(frame))
            self._time_field.setText(str(self._data.time()[frame]))
            # Update scene
            image = self._video_frames[self._video_iframe]
            image = QtGui.QImage(
                image, nx, ny, QtGui.QImage.Format.Format_ARGB32)
            self._video_pixmap.convertFromImage(image)
            self._video_item.setPixmap(self._video_pixmap)   # qt5
        # Update graph area
        self._graph_area.set_position(self._video_iframe)

    def action_start_timer(self):
        """
        Starts the timer.
        """
        self._timer_paused = False
        if self._data is not None:
            self._timer.start()
            self._play_button.setIcon(self._play_icon_pause)

    def action_stop_timer(self):
        self._timer_paused = False
        self._timer.stop()
        self._play_button.setIcon(self._play_icon_play)

    def closeEvent(self, event=None):
        # Called when window is closed.
        self.save_config()
        if event:
            event.accept()

    def create_menu(self):
        """
        Creates this widget's menu.
        """
        self._menu = self.menuBar()
        # File menu
        self._menu_file = self._menu.addMenu('&File')
        # File > Open
        self._tool_open = QtGui.QAction('&Open', self)
        self._tool_open.setShortcut('Ctrl+O')
        self._tool_open.setStatusTip('Open a DataLog for inspection.')
        self._tool_open.setIcon(QtGui.QIcon.fromTheme('document-open'))
        self._tool_open.triggered.connect(self.action_open)
        self._menu_file.addAction(self._tool_open)
        # File > ----
        self._menu_file.addSeparator()
        # File > Recent files
        self._recent_file_tools = []
        for i in range(N_RECENT_FILES):
            tool = QtGui.QAction(self, visible=False)
            tool.triggered.connect(self.action_recent_file)
            self._recent_file_tools.append(tool)
            self._menu_file.addAction(tool)
        # File > ----
        self._menu_file.addSeparator()
        # File > Quit
        self._tool_exit = QtGui.QAction('&Quit', self)
        self._tool_exit.setShortcut('Ctrl+Q')
        self._tool_exit.setStatusTip('Exit application.')
        self._tool_exit.setIcon(QtGui.QIcon.fromTheme('application-exit'))
        self._tool_exit.triggered.connect(self.close)
        self._menu_file.addAction(self._tool_exit)
        # Data menu
        self._menu_data = self._menu.addMenu('&Data')
        # Data > Extract frame
        self._tool_exframe = QtGui.QAction('Extract &frame', self)
        self._tool_exframe.setStatusTip(
            'Extract the current frame as csv file.')
        self._tool_exframe.triggered.connect(self.action_extract_frame)
        self._menu_data.addAction(self._tool_exframe)
        # Data > Extract graphs
        self._tool_exgraph = QtGui.QAction('Extract &graphs', self)
        self._tool_exgraph.setStatusTip('Extract the current graphs.')
        self._tool_exgraph.triggered.connect(self.action_extract_graphs)
        self._menu_data.addAction(self._tool_exgraph)
        # Data > ----
        self._menu_data.addSeparator()
        # Data > Extract frame as image
        self._tool_imgframe = QtGui.QAction('Save frame as &image', self)
        self._tool_imgframe.setStatusTip(
            'Save the current frame as an image file.')
        self._tool_imgframe.triggered.connect(self.action_extract_frame_image)
        self._menu_data.addAction(self._tool_imgframe)
        # Data > Extract colormap as image
        self._tool_imgcolor = QtGui.QAction('Save &colormap as image', self)
        self._tool_imgcolor.setStatusTip('Save the colormap as an image file.')
        self._tool_imgcolor.triggered.connect(
            self.action_extract_colormap_image)
        self._menu_data.addAction(self._tool_imgcolor)
        # Help menu
        self._menu_help = self._menu.addMenu('&Help')
        # Help > About
        self._tool_about = QtGui.QAction('&About', self)
        self._tool_about.setStatusTip('View information about this program.')
        self._tool_about.triggered.connect(self.action_about)
        self._menu_help.addAction(self._tool_about)
        self._tool_license = QtGui.QAction('&License', self)
        self._tool_license.setStatusTip('View this program\'s license info.')
        self._tool_license.triggered.connect(self.action_license)
        self._menu_help.addAction(self._tool_license)

    def display_exception(self):
        """ Displays the last exception in a messagebox. """
        QtWidgets.QMessageBox.warning(
            self, TITLE,
            '<h1>An error has occurred.</h1>'
            f'<pre>{traceback.format_exc()}</pre>')

    def event_colormap_selected(self):
        """ Colormap is selected by user. """
        self.action_set_colormap(str(self._colormap_select.currentText()))

    def event_graph_mouse_move(self, x, y):
        """
        Graph cursur moved: Display the current cursor position on the graph
        scene in the status bar.
        """
        self._label_cursor.setText(f'({x:< 1.6g}, {y:< 1.6g})')

    def event_rate_changed(self, e=None):
        """ User changed frame interval. """
        self._timer_interval = int(self._rate_field.text())
        self._timer.setInterval(self._timer_interval)

    def event_variable_selected(self):
        """ Variable is selected by user. """
        self._variable = self._variable_select.currentText()
        if self._data is not None:
            self.action_set_variable(self._variable)

    def event_video_single_click(self, x, y):
        """ Video clicked: Add a graph at the location of the click. """
        self._graph_area.graph(self._variable, x, y)

    def event_video_double_click(self, x, y):
        """
        Video double clicked: Add a frozen graph at the location of the click
        """
        self._graph_area.graph(self._variable, x, y)
        self._graph_area.freeze()

    def event_video_mouse_move(self, x, y):
        """
        Video cursur moved: Display the current cursor position on the video
        scene in the status bar.
        """
        # Cursor position is in scene coordinates, so already matches datablock
        # dimensions!
        if self._data is not None:
            try:
                z = self._data.get2d(self._variable)[self._video_iframe, y, x]
                z = '{:< 1.6g}'.format(z)
            except IndexError:
                z = '?'
            self._label_cursor.setText(f'({x:< 1.6g}, {y:< 1.6g}, {z}')

    def keyPressEvent(self, e):
        # A key has been pressed

        if e.key() == Qt.Key.Key_Space:
            self.action_start_stop()

    def load_config(self):
        """
        Loads the user configuration from an ini file.
        """
        # Read ini file
        config = configparser.RawConfigParser()
        try:
            config.read(os.path.expanduser(SETTINGS_FILE))
        except configparser.ParsingError:
            # Partially read config causes all sorts of errors, so discard
            config = configparser.RawConfigParser()

        # Window dimensions and location
        if config.has_section('window'):
            g = self.geometry()

            def getor(name, alt):
                if config.has_option('window', name):
                    return int(config.get('window', name))
                return alt
            x = getor('x', g.x())
            y = getor('y', g.y())
            w = getor('w', g.width())
            h = getor('h', g.height())
            self.setGeometry(x, y, w, h)
        # Splitter position
        if config.has_section('splitter'):
            if config.has_option('splitter', 's1') and config.has_option(
                    'splitter', 's2'):
                s = [int(config.get('splitter', x)) for x in ('s1', 's2')]
                self._central_widget.setSizes(s)

        # Current files, directory, etc
        if config.has_section('files'):
            if config.has_option('files', 'path'):
                path = config.get('files', 'path')
                if os.path.isdir(path):
                    self._path = path
            self._recent_files = []
            for i in range(0, N_RECENT_FILES):
                opt = 'recent_' + str(i)
                if config.has_option('files', opt):
                    fname = config.get('files', opt)
                    if os.path.isfile(fname):
                        self._recent_files.append(fname)
            self.update_recent_files_menu()

        # Selected colormap
        if config.has_section('video'):
            if config.has_option('video', 'color_map'):
                cmap = config.get('video', 'color_map')
                self.action_set_colormap(cmap)
            if config.has_option('video', 'interval'):
                ival = config.get('video', 'interval')
                if ival:
                    ival = int(ival)
                    if ival > 0:
                        self._timer_interval = ival

    def load_data_file(self, fname):
        """
        Attempts to load the given data block 2d file.
        """
        self._timer.stop()
        # Fix path
        fname = os.path.abspath(str(fname))
        self._path = os.path.dirname(fname)
        # Try loading file.
        self.statusBar().showMessage('Loading ' + str(fname))
        n = 1000000
        pd = QtWidgets.QProgressDialog('Loading data file...', 'Cancel', 0, n)
        pd.setWindowModality(Qt.WindowModality.WindowModal)
        pd.setValue(0)

        class Reporter(myokit.ProgressReporter):
            def __init__(self, pd):
                self._pd = pd

            def enter(self, msg=None):
                pass

            def exit(self):
                pass

            def update(self, f):
                self._pd.setValue((int)(n * f))
                return not self._pd.wasCanceled()

        reporter = Reporter(pd)
        try:
            data = myokit.DataBlock2d.load(fname, progress=reporter)
            del reporter
        except myokit.DataBlockReadError as e:
            pd.reset()
            self.statusBar().showMessage('Load failed.')
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>Unable to read file.</h1>'
                f'<p>The given filename <code>{fname}</code>'
                ' could not be read as a <code>myokit.DataBlock2d</code>.</p>'
                f'<p>{e}</p>'
            )
            return
        except Exception:
            pd.reset()
            self.statusBar().showMessage('Load failed.')
            self.display_exception()
            return
        finally:
            if pd.wasCanceled():
                self.statusBar().showMessage('Load canceled.')
                return

        # Don't load empty files
        if data.len2d() < 1:
            self.statusBar().showMessage('Load failed: empty file.')
            QtWidgets.QMessageBox.warning(
                self, TITLE,
                '<h1>Unable to read file.</h1>'
                f'<p>The given filename <code>{fname}</code>'
                ' does not contain any 2d data.</p>')
            return

        # File loaded okay
        self.statusBar().showMessage('File loaded succesfully.')
        self._file = fname
        self._data = data
        self.update_window_title()

        # Update recent file list
        try:
            # Remove fname from recent files list
            i = self._recent_files.index(fname)
            self._recent_files = \
                self._recent_files[:i] + self._recent_files[i + 1:]
        except ValueError:
            pass
        self._recent_files.insert(0, fname)
        self._recent_files = self._recent_files[:N_RECENT_FILES]
        self.update_recent_files_menu()

        # Update video scene
        nt, ny, nx = self._data.shape()
        self._video_scene.clear()
        self._video_scene.resize(nx, ny)
        self._video_view.resizeEvent()
        self._video_iframe = 0

        # Add empty video item to video scene
        self._video_pixmap = QtGui.QPixmap(nx, ny)
        self._video_item = QtWidgets.QGraphicsPixmapItem(self._video_pixmap)
        self._video_scene.addItem(self._video_item)

        # Update colormap scene
        nx, ny = self._colorbar_width, self._colorbar_height
        self._colorbar_scene.clear()
        self._colorbar_scene.resize(nx, ny)
        self._colorbar_view.resizeEvent()

        # Add empty colormap item to colormap scene
        self._colorbar_pixmap = QtGui.QPixmap(nx, ny)
        self._colorbar_item = QtWidgets.QGraphicsPixmapItem(
            self._colorbar_pixmap)
        self._colorbar_scene.addItem(self._colorbar_item)

        # Move slider to correct position
        self._slider.setMaximum(nt)
        self._slider.setPageStep(int(nt / 20))

        # Update controls
        for i in range(self._variable_select.count(), 0, -1):
            self._variable_select.removeItem(i - 1)
        variable = None
        for name in data.keys2d():
            self._variable_select.addItem(name)
            if variable is None or name in ('membrane.V', 'membrane.v'):
                variable = name
        self.action_set_variable(variable)
        self.action_set_colormap(self._colormap)

        # Update graph area
        self._graph_area.set_data(self._data)

    def _resize_started(self, e=None):
        """
        Called when the video scene is resized, delays video updates by a few
        ms.
        """
        self._resize_timer.start(300)
        self.action_pause_timer()

    def _resize_timeout(self, e=None):
        """
        Called a few ms after a resize.
        """
        self._resize_timer.stop()
        self.action_depause_timer()

    def save_config(self):
        """
        Saves the user configuration to an ini file.
        """
        config = configparser.RawConfigParser()

        # Window dimensions and location
        config.add_section('window')
        g = self.geometry()
        config.set('window', 'x', str(g.x()))
        config.set('window', 'y', str(g.y()))
        config.set('window', 'w', str(g.width()))
        config.set('window', 'h', str(g.height()))

        # Splitter position
        config.add_section('splitter')
        s = self._central_widget.sizes()
        config.set('splitter', 's1', str(s[0]))
        config.set('splitter', 's2', str(s[1]))

        # Current files, directory, etc
        config.add_section('files')
        config.set('files', 'path', self._path)
        for k, fname in enumerate(self._recent_files):
            config.set('files', 'recent_' + str(k), fname)

        # Current control values
        config.add_section('video')
        config.set('video', 'color_map', self._colormap)
        config.set('video', 'interval', self._timer_interval)

        # Write configuration to ini file
        inifile = os.path.expanduser(SETTINGS_FILE)
        with open(inifile, 'w') as configfile:
            config.write(configfile)

    def update_recent_files_menu(self):
        """
        Updates the recent files menu.
        """
        for k, fname in enumerate(self._recent_files):
            t = self._recent_file_tools[k]
            t.setText(f'{k + 1}. {os.path.basename(fname)}')
            t.setData(fname)
            t.setVisible(True)
        for i in range(len(self._recent_files), N_RECENT_FILES):
            self._recent_file_tools[i].setVisible(False)

    def update_window_title(self):
        """
        Sets this window's title based on the current state.
        """
        title = f'{TITLE} {myokit.__version__}'
        if self._file:
            title = f'{os.path.basename(self._file)} - {title}'
        self.setWindowTitle(title)


class VideoScene(QtWidgets.QGraphicsScene):
    """
    Color data display scene.

    Note that, despite the name, this item does not manage the conversion from
    data to the image. The actual drawing happens by calling ``setPixmap`` on a
    ``QGraphicsPixmapItem`` that gets added to this scene.

    See :meth:`DataBlockViewer.action_set_variable()` and
    :meth:`DataBlockViewer.action_set_frame()`.
    """
    # Signals
    # Somebody moved the mouse
    # Attributes: cell x, cell y
    mouse_moved = QtCore.Signal(int, int)
    # Single click
    # Attributes: cell x, cell y
    single_click = QtCore.Signal(float, float)
    # Double click!
    # Attributes: cell x, cell y
    double_click = QtCore.Signal(float, float)

    def __init__(self, *args):
        super().__init__(*args)
        self.setBackgroundBrush(QtGui.QColor(192, 192, 192))
        self._w = None
        self._h = None
        self._p = None
        self.resize(1, 1)

    def mousePressEvent(self, event):
        # Single-click event
        if event.button() == Qt.MouseButton.LeftButton:
            if event.modifiers() == Qt.KeyboardModifier.NoModifier:
                p = event.scenePos()
                x, y = int(p.x()), int(p.y())
                if x >= 0 and x < self._w and y >= 0 and y < self._h:
                    self.single_click.emit(x, y)
                    return

    def mouseDoubleClickEvent(self, event):
        # Double-click event
        if event.button() == Qt.MouseButton.LeftButton:
            if event.modifiers() == Qt.KeyboardModifier.NoModifier:
                p = event.scenePos()
                x, y = int(p.x()), int(p.y())
                if x >= 0 and x < self._w and y >= 0 and y < self._h:
                    self.double_click.emit(x, y)
                    return

    def mouseMoveEvent(self, event):
        # The mouse has moved!
        p = event.scenePos()
        x, y = int(p.x()), int(p.y())
        self.mouse_moved.emit(x, y)

    def resize(self, w, h):
        """ Resize the scene to match the given dimensions. """
        self._w = float(w)
        self._h = float(h)
        self.setSceneRect(0, 0, self._w, self._h)


class VideoView(QtWidgets.QGraphicsView):
    """
    Views a color data scene.
    """
    # Signals
    # The view was resized
    resize_event = QtCore.Signal()

    def __init__(self, scene):
        super().__init__(scene)
        # Disable scrollbars
        self.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        # Set rendering hints
        self.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        self.setViewportUpdateMode(
            QtWidgets.QGraphicsView.ViewportUpdateMode.BoundingRectViewportUpdate)  # noqa

        # Fit scene rect in view
        self.fitInView(self.sceneRect(), keepAspect=True)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)

        # Delayed resizing
        self._resize_timer = QtCore.QTimer()
        self._resize_timer.timeout.connect(self._resize_timeout)

        # No focus (no keyboard, no border)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)

    def fitInView(self, rect, keepAspect=False):
        """
        For some reason, Qt has a stupid bug in it that gives the scene a
        (hardcoded) margin of 2px. To remove, this is a re-implementation of
        the fitInView method, loosely based on the original C.

        https://bugreports.qt-project.org/browse/QTBUG-11945
        """
        # Reset the view scale
        unity = self.transform().mapRect(QtCore.QRectF(0, 0, 1, 1))
        w, h = max(1, unity.width()), max(1, unity.height())
        self.scale(1. / w, 1 / h)

        # Find the ideal scaling ratio
        viewRect = self.viewport().rect()
        sceneRect = self.transform().mapRect(rect)
        xr = viewRect.width() / sceneRect.width()
        yr = viewRect.height() / sceneRect.height()
        if keepAspect:
            xr = min(xr, yr)

        # Scale and center
        self.scale(xr, yr)
        self.centerOn(rect.center())

    def resizeEvent(self, event=None):
        # Called when the view is resized.

        # Tell others a resize is happening
        # (This is used to pause the video playback)
        self.resize_event.emit()
        # Schedule delayed (and grouped) resize action
        self._resize_timer.start(100)

    def _resize_timeout(self, event=None):
        """
        Called a few ms after time out.
        """
        self._resize_timer.stop()
        self.fitInView(self.sceneRect(), keepAspect=True)


class GraphArea(QtWidgets.QWidget):
    """
    Area that can draw several graphs.
    """
    # Signals
    # Somebody moved the mouse
    # Attributes: cell x, cell y
    mouse_moved = QtCore.Signal(float, float)

    def __init__(self):
        super().__init__()
        # DataBlock 2d, and its time vector
        self._data = None
        self._time = None

        # Index (x, y, variable) of temporary graph (if any)
        self._temp_index = None
        self._temp_path = None

        # Map from indices to paths for all frozen graphs
        self._frozen = collections.OrderedDict()

        # Last variable used in temp or frozen graph: used to scale mouse y
        # coordinate
        self._last_variable = None

        # Scaling per variable
        self._scaling = {}

        # Time scaled to fit
        self._time_scaled = None
        self._tmin = 0
        self._tmax = 1
        self._trange = self._tmax - self._tmin

        # Current position in time
        self._position = self._tmin

        # Colors for drawing
        self._color_temp = Qt.GlobalColor.black
        self._color_cycle = [
            Qt.GlobalColor.red,
            #Qt.GlobalColor.green,
            Qt.GlobalColor.blue,
            #Qt.GlobalColor.cyan,
            Qt.GlobalColor.magenta,
            #Qt.GlobalColor.yellow,
            Qt.GlobalColor.darkRed,
            Qt.GlobalColor.darkGreen,
            Qt.GlobalColor.darkBlue,
            Qt.GlobalColor.darkCyan,
            Qt.GlobalColor.darkMagenta,
            Qt.GlobalColor.darkYellow,
        ]

        # Scaling factors from pixels to normalized (0, 1) coordinates. Updated
        # after every resize.
        self._sw = 1.0
        self._sh = 1.0

    def clear(self):
        """
        Removes all graphs from the widget.
        """
        self._frozen = collections.OrderedDict()
        self._scaling = {}
        self._temp_path = self._temp_index = None
        self._last_variable = None
        self.update()

    def freeze(self):
        """
        Adds the temporary graph to the set of frozen graphs.
        """
        if self._temp_index and self._temp_path:
            self._frozen[self._temp_index] = self._temp_path
        self._temp_index = self._temp_path = None
        self.update()

    def graph(self, variable, x, y):
        """
        Adds temporary graph to this widget.
        """
        if self._data is None:
            return

        # Create index, check for duplicates
        variable = self._last_variable = str(variable)
        x, y = int(x), int(y)
        index = (x, y, variable)
        if index == self._temp_index:
            return
        if index in self._frozen:
            self._temp_index = self._temp_path = None
            self.update()
            return

        # Get scaling info
        try:
            ymin, ymax = self._scaling[variable]
        except KeyError:
            data = self._data.get2d(variable)
            ymin = np.min(data)
            ymax = np.max(data)
            d = ymax - ymin
            ymin -= 0.05 * d
            ymax += 0.05 * d
            if ymin == ymax:
                ymin -= 1
                ymax += 1
            self._scaling[variable] = (ymin, ymax)

        # Create path, using real time and scaled y data
        xx = iter(self._time_scaled)
        yy = (self._data.trace(variable, x, y) - ymin) / (ymax - ymin)
        yy = iter(1 - yy)
        path = QtGui.QPainterPath()
        x, y = next(xx), next(yy)
        path.moveTo(x, y)
        for i in range(1, len(self._time)):
            x, y = next(xx), next(yy)
            path.lineTo(x, y)
        self._temp_index = index
        self._temp_path = path

        # Update!
        self.update()

    def log(self):
        """
        Returns a myokit DataLog containing the data currently displayed in the
        graph area.
        """
        d = myokit.DataLog()
        if self._data is not None:
            d['engine.time'] = self._data.time()
            for index in self._frozen.keys():
                x, y, variable = index
                d[variable, x, y] = self._data.trace(variable, x, y)
            if self._temp_index:
                x, y, variable = self._temp_index
                d[variable, x, y] = self._data.trace(variable, x, y)
        return d

    def minimumSizeHint(self):
        """
        Returns a minimum size.
        """
        return QtCore.QSize(250, 5)

    def mouseMoveEvent(self, event):
        """
        Trigger mouse moved event with graph coordinates.
        """
        if self._last_variable is None:
            return

        # Get normalized x, y coordinates ([0, 1])
        p = event.pos()
        x = float(p.x()) * self._sw
        y = 1 - float(p.y()) * self._sh

        # Scale x-axis according to time
        x = self._tmin + x * self._trange

        # Scale y-axis according to last shown variable
        ymin, ymax = self._scaling[self._last_variable]
        y = ymin + y * (ymax - ymin)

        # Emit event
        self.mouse_moved.emit(x, y)

    def paintEvent(self, event):
        """
        Draws all the graphs.
        """
        if self._data is None:
            return

        # Create painter
        painter = QtGui.QPainter()
        painter.begin(self)

        # Fill background
        painter.fillRect(self.rect(), QtGui.QBrush(Qt.GlobalColor.white))

        # Create coordinate system for graphs
        painter.scale(self.width(), self.height())
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)

        # Create pen
        pen = QtGui.QPen()
        pen.setWidth(0)

        # Draw frozen graphs
        colors = iter(self._color_cycle)
        for path in self._frozen.values():
            try:
                pen.setColor(next(colors))
            except StopIteration:
                colors = iter(self._color_cycle)
                pen.setColor(next(colors))
            painter.setPen(pen)
            painter.drawPath(path)

        # Draw temp graph
        if self._temp_path:
            pen.setColor(self._color_temp)
            painter.setPen(pen)
            painter.drawPath(self._temp_path)

        # Show time indicator
        pen.setColor(Qt.GlobalColor.red)
        painter.setPen(pen)
        t = (self._position - self._tmin) / self._trange
        painter.drawLine(QtCore.QLineF(t, 0, t, 1))

        # Finish
        painter.end()

    def resizeEvent(self, e=None):
        """
        Resized.
        """
        s = self.size()
        w, h = s.width(), s.height()
        self._sw = 1.0 / w if w > 0 else 1
        self._sh = 1.0 / h if h > 0 else 1

    def set_data(self, data):
        """
        Passes in the DataBlock2d this graph area extracts its data from.
        """
        self.clear()
        self._data = data
        self._time = data.time()

        self._tmin = self._time[0]
        self._tmax = self._time[-1]
        self._trange = self._tmax - self._tmin
        tpad = 0.01 * self._trange
        self._tmin -= tpad
        self._tmax += tpad
        self._trange += 2 * tpad
        self._time_scaled = (self._time - self._tmin) / self._trange

        self._position = self._tmin

    def set_position(self, pos):
        """
        Sets the position of the time indicator.
        """
        if self._data is not None:
            self._position = self._time[int(pos)]
            self.update()

    def sizeHint(self):
        """
        Returns a size suggestion.
        """
        return QtCore.QSize(250, 250)

    def sizePolicy(self):
        """
        Tells Qt that this widget shout expand.
        """
        return QtCore.QSizePolicy.Expanding


class AutoFloatField(QtWidgets.QLineEdit):
    """
    A QLineEdit that requires floats as input, but will show "(Auto)" when a
    non-float is entered and return ``None`` as its value.
    """
    def __init__(self, parent=None):
        super().__init__(parent)

        # Colors
        self._color_ok = QtGui.QTextCharFormat().foreground().color().getRgb()
        self._color_auto = (127, 127, 127)
        self._auto_text = '(Auto)'

        # Show text
        self.editingFinished.connect(self._autofy)
        self._autofy()

    def _autofy(self):
        """ Put (Auto) if not valid """
        if self.value() is None:
            self.setText(self._auto_text)
            c = self._color_auto
        else:
            c = self._color_ok
        self.setStyleSheet(f'color: rgb({c[0]}, {c[1]}, {c[2]})')

    def focusInEvent(self, event):
        super().focusInEvent(event)
        QtCore.QTimer.singleShot(0, self.selectAll)

    def value(self):
        try:
            return float(super().text())
        except ValueError:
            return None




#
# System information class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import importlib
import os
import platform
import sys

import myokit


def system(live_printing=False):
    """
    Returns a (long) string with system information.

    If ``live_printing`` is set to ``True``, no string is returned but the
    results are printed to screen as they come in.
    """

    # Print directly to screen or store in array
    if not live_printing:
        out = []
    else:
        # Create fake list interface that just prints
        class Out:
            def append(self, x):
                print(x)

            def extend(self, xs):
                print('\n'.join(xs))

        out = Out()

    # Basic system information
    out.append('== System information ==')
    out.append('Myokit: ' + myokit.version(raw=True))
    ver = iter(sys.version.splitlines())
    out.append('Python: ' + next(ver))
    out.extend([' ' * 8 + v for v in ver])
    out.append(
        'OS: ' + platform.system()
        + ' (' + sys.platform + ', ' + os.name + ')')
    out.append('    ' + platform.platform())
    out.append('')

    # Python requirements
    out.append('== Python requirements ==')
    out.append('NumPy: ' + _module_version('numpy'))
    out.append('SciPy: ' + _module_version('scipy'))
    out.append('Matplotlib: ' + _module_version('matplotlib'))
    out.append('ConfigParser: ' + _module_version('configparser'))
    out.append('Setuptools: ' + _module_version('setuptools'))
    out.append('')

    # Python extras
    out.append('== Python extras ==')
    out.append('SymPy: ' + _module_version('sympy'))
    out.append('MoviePy: ' + _module_version('moviepy'))
    out.append('')

    # GUI toolkits
    out.append('== GUI ==')

    try:    # pragma: no cover
        from PyQt6.QtCore import QT_VERSION_STR
        out.append('PyQt6: ' + QT_VERSION_STR)
        out.append('  Sip: ' + _module_version('sip'))
        del QT_VERSION_STR
    except ImportError:
        out.append('PyQt6: Not found')

    try:    # pragma: no cover
        from PyQt5.QtCore import QT_VERSION_STR
        out.append('PyQt5: ' + QT_VERSION_STR)
        out.append('  Sip: ' + _module_version('sip'))
        del QT_VERSION_STR
    except ImportError:
        out.append('PyQt5: Not found')

    out.append('PySide6: ' + _module_version('PySide6'))
    out.append('PySide2: ' + _module_version('PySide2'))
    out.append('')

    # Development tools
    out.append('== Development tools ==')
    out.append('Sphinx: ' + _module_version('sphinx'))
    out.append('Flake8: ' + _module_version('flake8'))
    out.append('')

    # Simulation tools / compilation
    out.append('== Simulation tools ==')
    compiler = myokit.Compiler.info()
    if compiler is None:    # pragma: no cover
        out.append('Compiler: NOT FOUND')
        out.append('Sundials: Compiler not found')
        out.append('OpenCL: Compiler not found')
    else:
        out.append('Compiler: ' + compiler)
        out.append('Sundials: ' + (myokit.Sundials.version() or 'Not found'))

        opencl = myokit.OpenCL()
        if not opencl.supported():  # pragma: no cover
            out.append('OpenCL: No OpenCL support detected.')
        else:   # pragma: no cover
            devices = []
            for p in opencl.info().platforms:
                for d in p.devices:
                    devices.append(d.name + ' on ' + p.name)
            out.append('OpenCL: ' + str(len(devices)) + ' device(s) found')
            for d in devices:
                out.append('  ' + d)
            out.append('  Use `python -m myokit opencl` for more information.')

    if not live_printing:
        return '\n'.join(out)


def _module_version(module):
    """
    Checks if the given ``module`` is present, and attempts to return a string
    containing its version. If the module can't be found 'Not found' is
    returned. If the module is found, but the  version can't be detected, 'OK'
    is returned.
    """
    try:
        m = importlib.import_module(module)
    except ImportError:     # pragma: no cover
        return 'Not found'

    # NumPy, SciPy, SymPy
    try:
        return str(m.version.version)
    except AttributeError:  # pragma: no cover
        pass

    # Matplotlib & others
    try:
        return str(m.__version__)
    except AttributeError:  # pragma: no cover
        pass

    # Setuptools
    try:
        return str(m.version.__version__)
    except AttributeError:  # pragma: no cover
        pass

    # Not found
    return 'OK'



#!/usr/bin/env python3
#
# Tests the OpenCL info class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit

from myokit.tests import TemporaryDirectory, OpenCL_FOUND


@unittest.skipIf(not OpenCL_FOUND, 'OpenCL not found on this system.')
class OpenCLTest(unittest.TestCase):
    """
    Tests the OpenCL info class.
    """

    def test_available(self):
        # Tested as condition to get in, so must be true
        self.assertTrue(myokit.OpenCL.available())

    def test_current_info(self):
        # Tests the method to query the current device
        self.assertIsInstance(
            myokit.OpenCL.current_info(), myokit.OpenCLPlatformInfo)
        self.assertIsInstance(myokit.OpenCL.current_info(True), str)

    def test_info(self):
        # Tests the method to query the current device
        self.assertIsInstance(myokit.OpenCL.info(), myokit.OpenCLInfo)
        self.assertIsInstance(myokit.OpenCL.info(True), str)

    def test_load_save_selection(self):
        # Tests the load_selection method
        import myokit._sim.opencl
        org_name = myokit._sim.opencl.SETTINGS_FILE
        try:
            with TemporaryDirectory() as d:
                fname = d.path('opencl-temp.ini')
                myokit._sim.opencl.SETTINGS_FILE = fname

                # Save None and None
                myokit.OpenCL.save_selection(None, None)
                platform, device = myokit.OpenCL.load_selection()
                self.assertIsNone(platform)
                self.assertIsNone(device)

                myokit.OpenCL.save_selection(None, 'bert')
                platform, device = myokit.OpenCL.load_selection()
                self.assertIsNone(platform)
                self.assertEqual(device, 'bert')

                myokit.OpenCL.save_selection('ernie', None)
                platform, device = myokit.OpenCL.load_selection()
                self.assertEqual(platform, 'ernie')
                self.assertIsNone(device)

                myokit.OpenCL.save_selection('ernie', 'bert')
                platform, device = myokit.OpenCL.load_selection()
                self.assertEqual(platform, 'ernie')
                self.assertEqual(device, 'bert')

        finally:
            myokit._sim.opencl.SETTINGS_FILE = org_name

    def test_selection_info(self):
        # Tests getting info about devices, specifically for device selection
        info = myokit.OpenCL.selection_info()
        self.assertIsInstance(info, list)
        self.assertGreater(len(info), 0)
        item = info[0]
        self.assertEqual(len(item), 3)
        platform, device, specs = item
        self.assertIsInstance(platform, str)
        self.assertIsInstance(device, str)
        self.assertIsInstance(specs, str)

    def test_supported(self):
        # Tested as condition to get in, so must be true
        self.assertTrue(myokit.OpenCL.supported())

    def test_hidden(self):
        # Tests some private methods of the opencl module
        import myokit._sim.opencl as o

        # Test bytesize
        self.assertEqual(o.bytesize(3), '3 B')
        self.assertEqual(o.bytesize(1023), '1023 B')
        self.assertEqual(o.bytesize(1024), '1 KB')
        self.assertEqual(o.bytesize(2048), '2 KB')
        self.assertEqual(o.bytesize(1048575)[-2:], 'KB')
        self.assertEqual(o.bytesize(1048576), '1 MB')
        self.assertEqual(o.bytesize(1073741823)[-2:], 'MB')
        self.assertEqual(o.bytesize(1073741824), '1 GB')

        # Test clockspeed
        self.assertEqual(o.clockspeed(100), '100 MHz')
        self.assertEqual(o.clockspeed(999), '999 MHz')
        self.assertEqual(o.clockspeed(1000), '1 GHz')
        self.assertEqual(o.clockspeed(1200), '1.2 GHz')


if __name__ == '__main__':
    unittest.main()



#
# Provides EasyML (CARP/CARPentry) support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import EasyMLExporter
from ._ewriter import EasyMLExpressionWriter


# Importers

# Exporters
_exporters = {
    'easyml': EasyMLExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'easyml': EasyMLExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


#
# Language keywords: Not sure if these are all keywords
#
keywords = [
    'acos',
    'asin',
    'atan',
    'ceil',
    'cos',
    'fabs',
    'floor',
    'pow',
    'sin',
    'tan',
]



#!/usr/bin/env python3
#
# Tests the parsing module.
# See also test_io.py
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.units

from myokit.tests import DIR_DATA, TemporaryDirectory


class TokenizerTest(unittest.TestCase):
    """
    Tests the tokenizer class.
    """

    def test_tokenizer(self):
        # Test basic Tokenizer functionality.

        import myokit._parsing as p
        from myokit._parsing import Tokenizer
        s = Tokenizer('5')
        self.assertEqual(next(s), (p.INTEGER, '5', 1, 0))
        s = Tokenizer('3.0')
        self.assertEqual(next(s), (p.FLOAT, '3.0', 1, 0))
        s = Tokenizer('3.')
        self.assertEqual(next(s), (p.FLOAT, '3.', 1, 0))
        s = Tokenizer('3.E0')
        self.assertEqual(next(s), (p.FLOAT, '3.E0', 1, 0))
        s = Tokenizer('.30')
        self.assertEqual(next(s), (p.FLOAT, '.30', 1, 0))

        # Finished? Then should get a StopIteration
        self.assertEqual(s.peek()[0], p.EOL)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(s.peek()[0], p.EOF)
        self.assertEqual(next(s)[0], p.EOF)
        self.assertRaises(IndexError, s.peek)
        self.assertRaises(StopIteration, next, s)

        # Unknown token
        self.assertRaisesRegex(
            myokit.ParseError, 'invalid token', Tokenizer, '@')
        s = Tokenizer('x  @')
        self.assertRaisesRegex(
            myokit.ParseError, 'invalid token', s.next)

        # Block-comment
        s = Tokenizer('"""Hello"""')
        self.assertEqual(next(s)[0], p.EOF)

        # Multi-line string
        s = Tokenizer('x: """Hello\nWo\trld"""')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertEqual(next(s)[0], p.COLON)
        self.assertEqual(s.peek()[0], p.TEXT)
        self.assertEqual(next(s)[1], 'Hello\nWo\trld')
        self.assertEqual(next(s)[0], p.EOL)
        s = Tokenizer('x: """Hello"""World')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertRaisesRegex(
            myokit.ParseError, 'after closing of multi-line string', next, s)
        s = Tokenizer('x: """He\nll\no"""World')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertRaisesRegex(
            myokit.ParseError, 'after closing of multi-line string', next, s)
        s = Tokenizer('x: """Hello\n')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertRaisesRegex(
            myokit.ParseError, 'Unclosed multi-line', s.next)

        # Indented is removed from multi-line string
        s = Tokenizer('    x: """Hello \n    world"""')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertEqual(next(s)[0], p.COLON)
        self.assertEqual(s.peek()[0], p.TEXT)
        self.assertEqual(s.peek()[1], 'Hello\nworld')
        s = Tokenizer('\tx: """Hello \n\tworld"""')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertEqual(next(s)[0], p.COLON)
        self.assertEqual(s.peek()[0], p.TEXT)
        self.assertEqual(s.peek()[1], 'Hello\nworld')

        # Empty lines
        s = Tokenizer('\n\n\nx: """Hello\nWorld"""\n\n\n')
        next(s)
        next(s)
        self.assertEqual(next(s)[1], 'Hello\nWorld')
        s = Tokenizer('\n\n\nx: """Hello\n\n\nWorld"""\n\n\n')
        next(s)
        next(s)
        self.assertEqual(next(s)[1], 'Hello\n\n\nWorld')

        # Line continuation
        s = Tokenizer('x: this is \\\nthe value of x')
        next(s)
        next(s)
        self.assertEqual(next(s)[1], 'this is the value of x')
        self.assertEqual(next(s)[0], p.EOL)

        s = Tokenizer('x: this is \\\nthe\\\n\n\n value of x')
        next(s)
        next(s)
        self.assertEqual(next(s)[1], 'this is the')
        self.assertEqual(next(s)[0], p.EOL)

        s = Tokenizer('x: this is \\\nthe\\\n\n\n value of x')
        next(s)
        next(s)
        self.assertEqual(next(s)[1], 'this is the')
        self.assertEqual(next(s)[0], p.EOL)

        # Comment doesn't end line continuation
        s = Tokenizer('x: this is \\\n#Hi mike\nthe value of x')
        next(s)
        next(s)
        self.assertEqual(next(s)[1], 'this is the value of x')

        # Line continuation must be last character on line
        s = Tokenizer('1 + \\3')
        next(s)
        self.assertRaisesRegex(
            myokit.ParseError, ' Backslash must be last', s.next)

        # But backslash is allowed in text (only seen as line cont. if last)
        s = Tokenizer('x: Hello\\michael\nHow are you')
        self.assertEqual(next(s)[0], p.META_NAME)
        self.assertEqual(next(s)[0], p.COLON)
        self.assertEqual(next(s)[1], 'Hello\\michael')

        # Brackets
        s = Tokenizer('x = (\n1 + \n\n2 + (\n3\n\n)\n\n) + 4')
        next(s)
        next(s)
        from myokit._parsing import parse_expression_stream
        e = parse_expression_stream(s)
        self.assertEqual(e.code(), '1 + 2 + 3 + 4')

        # Mismatched brackets
        s = Tokenizer('x = (1 + 2')
        self.assertEqual(next(s)[0], p.NAME)
        self.assertEqual(next(s)[0], p.EQUAL)
        self.assertEqual(next(s)[0], p.PAREN_OPEN)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.PLUS)
        self.assertRaisesRegex(
            myokit.ParseError, 'Parentheses mismatch', s.next)

        s = Tokenizer('x = 1 + 2)')
        self.assertEqual(next(s)[0], p.NAME)
        self.assertEqual(next(s)[0], p.EQUAL)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.PLUS)
        self.assertRaisesRegex(
            myokit.ParseError, 'Parentheses mismatch', s.next)

        # Test indenting (Tabs count as 8 spaces)
        s = '\n'.join([
            '1',
            '        2',
            '\t3',
            '        4',
            '5',
        ])
        s = Tokenizer(s)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(next(s)[0], p.INDENT)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(next(s)[0], p.DEDENT)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)

        # Test indenting error
        s = Tokenizer('\n'.join([
            '1',
            '    2',
            '  3',
        ]))
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(next(s)[0], p.INDENT)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertRaisesRegex(
            myokit.ParseError, 'Unexpected indenting level', s.next)

        # Line feed counts as a newline
        s = Tokenizer('123\f456')
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)
        self.assertEqual(next(s)[0], p.INTEGER)
        self.assertEqual(next(s)[0], p.EOL)


class PhasedParseTest(unittest.TestCase):
    """
    Tests several phases of parsing.
    """

    def test_segment_parsing(self):
        # Test :meth:`parse_model()`.
        from myokit._parsing import parse

        # Empty code --> error
        code = ''
        self.assertRaisesRegex(
            myokit.ParseError,
            r'expecting Segment header "\[\[segment_name]]"',
            parse, code)

        # Unexpected before model
        code = (
            '[[bert]]\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError,
            r'Expecting \[\[model]] or \[\[protocol]] or \[\[script]]',
            parse, code)

        # Unexpected before protocol
        code = (
            '[[model]]\n',
            '[c]\n',
            't = 0 bind time\n',
            '[[bert]]\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError,
            r'Expecting \[\[protocol]] or \[\[script]]',
            parse, code)

        # Unexpected before script
        code = (
            '[[model]]\n',
            '[c]\n',
            't = 0 bind time\n',
            '[[protocol]]\n',
            '[[bert]]\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError,
            r'Expecting \[\[script]]',
            parse, code)

        # Bad segment order (same as unexpected after protocol)
        code = (
            '[[protocol]]\n',
            '[[model]]\n',
            '[c]\n',
            't = 0 bind time\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[script]]', parse, code)

    def test_parse_model(self):
        # Test the parse_model method.
        from myokit._parsing import parse_model

        # Test simple
        code = (
            '[[model]]\n',
            '[c]\n',
            't = 0 bind time\n',
        )
        model = parse_model(code)
        self.assertIsInstance(model, myokit.Model)
        model = parse_model(''.join(code))
        self.assertIsInstance(model, myokit.Model)

        # Not a model
        code = (
            '[[muddle]]\n',
            '[c]\n',
            't = 0 bind time\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[model]]', parse_model, code)

    def test_parse_model_from_stream_error(self):
        # Quick error testing for :meth:`parse_model_from_stream`.
        from myokit._parsing import parse_model_from_stream
        from myokit._parsing import Tokenizer

        def p(code):
            return parse_model_from_stream(Tokenizer(iter(code)))

        code = (
            '[[model]]',
            '[c]',
            't = 0 bind time',
        )
        model = p(code)
        self.assertIsInstance(model, myokit.Model)

        # Not a model
        code = (
            '[[muddle]]',
            '[c]',
            't = 0 bind time',
        )
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[model]]', p, code)

        # Double meta-data value
        code = (
            '[[model]]',
            'ax: 1',
            'ax: 1',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Duplicate meta-data key', p, code)

        # Double initial values
        code = (
            '[[model]]',
            'a.x = 1',
            'a.x = 1',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Duplicate initial value', p, code)

        # Unused initial values
        code = (
            '[[model]]',
            'a.x = 1',
            '[c]',
            't = 0 bind time',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Unused initial value', p, code)

    def test_parse_user_function(self):
        # Test :meth:`parse_user_function()`.
        from myokit._parsing import parse_model as p

        # Test basics
        code = (
            '[[model]]',
            'michael(x) = 1 + sin(x)',
            '[c]',
            't = 0 bind time',
            'x = michael(12)',
        )
        p(code)

        code = (
            '[[model]]',
            'michael(x, y, z) = x + y / z',
            '[c]',
            't = 0 bind time',
            'x = michael(1,2,3)',
        )
        p(code)

        # Duplicate name
        code = (
            '[[model]]',
            'michael(x, y, z) = x + y / z',
            'michael(x, y, z) = x - y - z',
            '[c]',
            't = 0 bind time',
            'x = michael(1,2,3)',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'already defined', p, code)

    def test_block_comments(self):
        # Test block comments in model.
        from myokit._parsing import parse_model_from_stream
        from myokit._parsing import Tokenizer

        def p(code):
            return parse_model_from_stream(Tokenizer(iter(code)))

        # Block comments
        c1 = (
            '[[model]]',
            '[c]',
            't = 0 bind time',
        )
        m1 = p(c1)
        c2 = (
            '[[model]]',
            '"""This is a comment"""',
            '[c]',
            't = 0 bind time',
        )
        m2 = p(c2)
        c3 = (
            '[[model]]',
            '"""'
            'This is a long',
            'long',
            'long',
            'comment',
            '"""',
            '[c]',
            't = 0 bind time',
        )
        m3 = p(c3)
        self.assertEqual(m1.code(), m2.code(), m3.code())

    def test_parse_component(self):
        # Test parse_component(), uses parse_variable
        from myokit._parsing import parse_model as p

        # Test basics
        code = (
            '[[model]]',
            '[test]',
            't = 0',
            '    bind time',
            'x = 5 + b',
            '    b = 13',
            '    in [mV]',
            'desc: \"""',
            'This is a test component.',
            '\"""',
        )
        m = p(code)
        self.assertIn('test', m)
        self.assertEqual(
            m.get('test').meta['desc'], 'This is a test component.')

        # Test duplicate name
        code = (
            '[[model]]',
            '[test]',
            't = 0',
            '    bind time',
            '[test]',
            'x = 2',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Duplicate component name', p, code)

        # Test invalid name --> Handled by tokenizer

        # Test duplicate meta data key
        code = (
            '[[model]]',
            '[test]',
            'yes: no',
            't = 0',
            '    bind time',
            'yes: yes',
            '[test]',
            'x = 2',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Duplicate meta-data key', p, code)

    def test_parse_alias(self):
        # Test :meth:`parse_alias()`.
        from myokit._parsing import parse_model as p

        # Test basics
        code = (
            '[[model]]',
            '[a]',
            't = 10',
            '    bind time',
            '[b]',
            'use a.t as time',
            'b = time',
            '\"""',
        )
        m = p(code)
        self.assertEqual(m.get('b.b').eval(), 10)

        # Test bad variable name
        code = (
            '[[model]]',
            '[a]',
            't = 10',
            '    bind time',
            '[b]',
            'use t as time',
            'b = time',
            '\"""',
        )
        self.assertRaisesRegex(myokit.ParseError, 'fully qualified', p, code)

        # Test bad alias name
        code = (
            '[[model]]',
            '[a]',
            't = 10',
            '    bind time',
            '[b]',
            'use a.t as _flbt12',
            'b = time',
            '\"""',
        )
        self.assertRaisesRegex(myokit.ParseError, 'invalid token', p, code)

        # Unknown variable
        code = (
            '[[model]]',
            '[a]',
            't = 10',
            '    bind time',
            '[b]',
            'use a.x as time',
            'b = time',
            '\"""',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Variable not found', p, code)

    def test_parse_variable(self):
        # Test parse_variable(), uses parse_expression()
        from myokit._parsing import parse_model as p

        # Test basics
        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
        )
        t = p(s).get('x.t')
        self.assertEqual(t.name(), 't')
        self.assertEqual(t.qname(), 'x.t')
        self.assertEqual(t.unit(), None)
        self.assertEqual(t.binding(), 'time')

        s = (
            '[[model]]',
            '[x]',
            't = 0',
            '    bind time',
        )
        t = p(s).get('x.t')
        self.assertEqual(t.name(), 't')
        self.assertEqual(t.qname(), 'x.t')
        self.assertEqual(t.unit(), None)
        self.assertEqual(t.binding(), 'time')

        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x1 = 5'
        )
        v = p(s).get('x.x1')
        self.assertIsInstance(v, myokit.Variable)
        self.assertTrue(v.is_constant())
        self.assertTrue(v.is_literal())
        self.assertFalse(v.is_state())
        self.assertFalse(v.is_intermediary())
        self.assertEqual(v.unit(), None)
        self.assertEqual(v.rhs().unit(), None)

        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x2 = 5 [mV]'
        )
        v = p(s).get('x.x2')
        self.assertIsInstance(v, myokit.Variable)
        self.assertTrue(v.is_constant())
        self.assertTrue(v.is_literal())
        self.assertFalse(v.is_state())
        self.assertFalse(v.is_intermediary())
        self.assertEqual(v.unit(), None)
        self.assertEqual(v.rhs().unit(), myokit.parse_unit('mV'))

        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x3 = 5',
            '    in [mV]',
        )
        v = p(s).get('x.x3')
        self.assertIsInstance(v, myokit.Variable)
        self.assertTrue(v.is_constant())
        self.assertTrue(v.is_literal())
        self.assertFalse(v.is_state())
        self.assertFalse(v.is_intermediary())
        self.assertEqual(v.unit(), myokit.parse_unit('mV'))
        self.assertEqual(v.rhs().unit(), None)

        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x3 = 5 in [mV]',
        )
        v = p(s).get('x.x3')
        self.assertIsInstance(v, myokit.Variable)
        self.assertTrue(v.is_constant())
        self.assertTrue(v.is_literal())
        self.assertFalse(v.is_state())
        self.assertFalse(v.is_intermediary())
        self.assertEqual(v.unit(), myokit.parse_unit('mV'))
        self.assertEqual(v.rhs().unit(), None)

        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x4 = 5 [V] : This is x4',
        )
        v = p(s).get('x.x4')
        self.assertIsInstance(v, myokit.Variable)
        self.assertTrue(v.is_constant())
        self.assertTrue(v.is_literal())
        self.assertFalse(v.is_state())
        self.assertFalse(v.is_intermediary())
        self.assertEqual(v.unit(), None)
        self.assertEqual(v.rhs().unit(), myokit.units.V)
        self.assertEqual(v.meta['desc'], 'This is x4')

        code = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x = 5 label vvv',
        )
        x = p(code).get('x.x')
        self.assertIsInstance(x, myokit.Variable)
        self.assertEqual(x.label(), 'vvv')

        s = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x = 5',
            '    label vvv',
        )
        x = p(code).get('x.x')
        self.assertIsInstance(x, myokit.Variable)
        self.assertEqual(x.label(), 'vvv')

        # Illegal lhs
        code = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'sin(x) = 5',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'variable names or the dot', p, code)

        # Duplicate name
        code = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'x = 5',
            'x = 5',
        )
        self.assertRaisesRegex(myokit.ParseError, 'Duplicate var', p, code)

        # Missing initial value
        code = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            'dot(x) = 5',
        )
        self.assertRaisesRegex(myokit.ParseError, 'Missing initial', p, code)

        # Duplicate meta
        code = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            '    yes: really',
            '    yes: no',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Duplicate meta-data key', p, code)

        # Duplicate unit
        code = (
            '[[model]]',
            '[x]',
            't = 0 bind time',
            '    in [s]',
            '    in [s]',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'Duplicate variable unit', p, code)

    def test_parse_unit(self):
        # Test :meth:`parse_unit` and :meth:`parse_unit_string`.
        from myokit._parsing import parse_unit_string as p

        # Test dimensionless
        self.assertEqual(p('1'), myokit.units.dimensionless)

        # Test bare unit
        self.assertEqual(p('V'), myokit.units.Volt)

        # Test unit with quantifier
        self.assertEqual(p('mV'), myokit.units.Volt / 1000)

        # Test multiplied units
        self.assertEqual(p('V*A'), myokit.units.Volt * myokit.units.Ampere)
        self.assertEqual(p('J/s'), myokit.units.Watt)
        self.assertEqual(p('1/s'), 1 / myokit.units.second)

        # Test units with exponents (first unit)
        self.assertEqual(p('m^2'), myokit.units.meter ** 2)
        self.assertEqual(p('m^-1'), 1 / myokit.units.meter)
        # Exponents on remaining units
        self.assertEqual(
            p('s*m^2'), myokit.units.second * myokit.units.meter ** 2)
        self.assertEqual(
            p('s*m^-1'), myokit.units.second / myokit.units.meter)

        # Test units with multipliers
        self.assertEqual(p('m (123)'), myokit.units.meter * 123)

        # Test bad unit
        self.assertRaisesRegex(
            myokit.ParseError, 'Unit not recognized', p, 'michael')
        self.assertRaisesRegex(
            myokit.ParseError, 'Unit not recognized', p, 'kg/michael')
        self.assertRaisesRegex(
            myokit.ParseError, 'Unexpected token', p, '*2')
        self.assertRaisesRegex(
            myokit.ParseError, 'Invalid unit specification', p, '2')
        self.assertRaisesRegex(
            myokit.ParseError, 'Invalid unit multiplier', p, 'm (x)')

    def test_parse_protocol(self):
        # Test :meth:`parse_protocol()`.
        from myokit._parsing import parse_protocol

        # Test simple
        # Level   Start   Length  Period  Multiplier
        code = (
            '[[protocol]]\n',
        )
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)
        protocol = parse_protocol(''.join(code))
        self.assertIsInstance(protocol, myokit.Protocol)

        code = (
            '[[protocol]]\n',
            '0 0 1 0 0\n',
        )
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)

        # Funny numbers
        code = ('[[protocol]]\n', '-1 0 1 0 0\n')
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)
        code = ('[[protocol]]\n', '-1 0 1e3 0 0\n')
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)
        code = ('[[protocol]]\n', '-1 0 1e3 1000 +3\n')
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)
        code = ('[[protocol]]\n', '0.1 0 1 0 0\n')
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)
        code = ('[[protocol]]\n', '0.1e-2 0 1 0 0\n')
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)
        code = ('[[protocol]]\n', '.1e-2 0 1 0 0\n')
        protocol = parse_protocol(code)
        self.assertIsInstance(protocol, myokit.Protocol)

        # Not a protocol
        code = (
            '[[protcle]]\n',
            '0 0 1 0 0\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[protocol]]',
            parse_protocol, code)

        # Not a protocol (in stream)
        from myokit._parsing import parse_protocol_from_stream, Tokenizer
        stream = Tokenizer(iter(code))
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[protocol]]',
            parse_protocol_from_stream, stream)

        # Test using 'next'
        code = (
            '[[protocol]]\n',
            '00 0 100 0 0\n',
            '10 next 100 0 0\n',
            '20 next 100 0 0\n',
        )
        p1 = parse_protocol(code)
        code = (
            '[[protocol]]\n',
            '00 0 100 0 0\n',
            '10 100 100 0 0\n',
            '20 200 100 0 0\n',
        )
        p2 = parse_protocol(code)
        self.assertEqual(p1.code(), p2.code())

        # Test invalid use of next (after periodic event)
        code = (
            '[[protocol]]\n',
            '00 0 100 1000 0\n',
            '10 next 100 0 0\n',
        )
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'Invalid next', parse_protocol, code)

        # Simultaneous events
        code = (
            '[[protocol]]\n',
            '00 0 100 0 0\n',
            '10 0 100 0 0\n',
        )
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'same time', parse_protocol, code)
        code = (
            '[[protocol]]\n',
            '00 0 100 1000 0\n',
            '10 0 1100 0 0\n',
        )
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'same time', parse_protocol, code)

        # Wrong order
        code = (
            '[[protocol]]\n',
            '0 10 1 0 0\n',
            '1 0  1 0 0\n',
        )
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'chronological', parse_protocol, code)

        # Duration <= 0
        code = ('[[protocol]]\n', '0 10 -1 0 0\n')
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'Invalid duration',
            parse_protocol, code)
        code = ('[[protocol]]\n', '0 10 0 0 0\n')
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'Invalid duration',
            parse_protocol, code)

        # Negative period
        code = ('[[protocol]]\n', '0 10 1 -1 0\n')
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'Negative period',
            parse_protocol, code)

        # Negative multiplier
        # Multiplier for non-periodic event
        code = ('[[protocol]]\n', '0 10 1 100 -1\n')
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'Negative multiplier',
            parse_protocol, code)

        # Multiplier for non-periodic event
        code = ('[[protocol]]\n', '0 10 1 0 1\n')
        self.assertRaisesRegex(
            myokit.ProtocolParseError, 'Invalid multiplier',
            parse_protocol, code)

        # Found a bug when parsing this
        code = (
            '[[protocol]]\n',
            '-80     next    300     0       0\n',
            '-120    3196.0  50.0    0       0\n',
        )
        parse_protocol(code)

    def test_parse_script(self):
        # Test :meth:`parse_script()`.
        from myokit._parsing import parse_script

        # Test simple
        code = (
            '[[script]]',
        )
        script = parse_script(code)
        self.assertIsInstance(script, str)
        script = parse_script(''.join(code))
        self.assertIsInstance(script, str)

        code = (
            '[[script]]\n',
        )
        script = parse_script(code)
        self.assertIsInstance(script, str)

        code = (
            '[[script]]\n',
            'print("hi")',
        )
        script = parse_script(code)
        self.assertIsInstance(script, str)

        code = (
            '[[script]]\n',
            'print("hi")\n',
        )
        script = parse_script(code)
        self.assertIsInstance(script, str)

        # Not a script
        code = (
            '[[hello]]\n',
            'print("hi")\n',
        )
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[script]]', parse_script, code)

        # Not a script, parsing from stream
        from myokit._parsing import Tokenizer, parse_script_from_stream
        raw = iter(code)
        stream = Tokenizer(raw)
        self.assertRaisesRegex(
            myokit.ParseError, r'Expecting \[\[script]]',
            parse_script_from_stream, stream, raw)

    def test_split(self):
        # Test split(), uses parse()
        from myokit._parsing import split

        # Test horrible scenario
        code = [
            '',
            '',
            '[[model]]',
            '',
            'Bad model',
            '',
            '',
            '',
            '[[protocol]]',
            '',
            '',
            '1 1 1 1 1',
            '',
            '',
            '4 4 4 4 4',
            'no'
            '',
            '',
            '[[script]]',
            '',
            '',
        ]
        m, p, x = split('\n'.join(code))
        self.assertEqual(m, '\n'.join(code[:8]))
        self.assertEqual(p, '\n'.join(code[8:17]))
        self.assertEqual(x, '\n'.join(code[17:-1]))

        # Scenario 2
        code[17] = '[[scrubbed]]'
        m, p, x = split('\n'.join(code))
        self.assertEqual(m, '\n'.join(code[:8]))
        self.assertEqual(p, '\n'.join(code[8:-1]))

    def test_unexpected_token(self):
        # Test :meth:`unexpected_token`.
        import myokit._parsing as p

        # code, text, line, char
        token = p.AND, 'and', 10, 5

        # Test with string expectation
        self.assertRaisesRegex(
            myokit.ParseError, 'expecting the spanish inquisition',
            p.unexpected_token, token, 'the spanish inquisition')

        # Test with one expected type
        self.assertRaisesRegex(
            myokit.ParseError, 'expecting End of line',
            p.unexpected_token, token, p.EOL)

        # Test with two expected type
        self.assertRaisesRegex(
            myokit.ParseError, 'expecting End of line or End of file',
            p.unexpected_token, token, [p.EOL, p.EOF])

        # Test with many expected type
        self.assertRaisesRegex(
            myokit.ParseError,
            r'expecting one of \[Plus "\+", Minus "-", Star "\*"]',
            p.unexpected_token, token, [p.PLUS, p.MINUS, p.STAR])

    def test_parse_expression(self):
        # Test parse_expression()
        from myokit import parse_expression as p
        from myokit import Number
        e = p('5')
        self.assertIsInstance(e, Number)
        self.assertEqual(e.eval(), 5.0)
        self.assertEqual(float(e), 5.0)
        e = p('5[m]')
        self.assertIsInstance(e, Number)
        self.assertEqual(e.eval(), 5.0)
        self.assertEqual(float(e), 5.0)
        self.assertEqual(e.unit(), myokit.units.m)
        e = p('5 [m/s]')
        self.assertIsInstance(e, Number)
        self.assertEqual(e.eval(), 5.0)
        self.assertEqual(float(e), 5.0)
        self.assertEqual(e.unit(), myokit.parse_unit('m/s'))
        self.assertEqual(e.unit(), myokit.units.m / myokit.units.s)
        e = p('+5')
        self.assertIsInstance(e, myokit.PrefixPlus)
        self.assertEqual(e.eval(), 5.0)
        e = p('++5')
        self.assertIsInstance(e, myokit.PrefixPlus)
        self.assertEqual(e.eval(), 5.0)
        e = p('-5')
        self.assertIsInstance(e, myokit.PrefixMinus)
        self.assertEqual(e.eval(), -5.0)
        e = p('--5')
        self.assertIsInstance(e, myokit.PrefixMinus)
        self.assertEqual(e.eval(), 5.0)
        e = p('5 + 2')
        self.assertIsInstance(e, myokit.Plus)
        self.assertEqual(e.eval(), 7)
        e = p('5 + 1 + 1')
        self.assertIsInstance(e, myokit.Plus)
        self.assertEqual(e.eval(), 7)
        e = p('5 - 2')
        self.assertIsInstance(e, myokit.Minus)
        self.assertEqual(e.eval(), 3)
        e = p('5 -- 2')
        self.assertIsInstance(e, myokit.Minus)
        self.assertEqual(e.eval(), 7)
        e = p('5 --+ 2')
        self.assertIsInstance(e, myokit.Minus)
        self.assertEqual(e.eval(), 7)
        e = p('5 --- 2')
        self.assertIsInstance(e, myokit.Minus)
        self.assertEqual(e.eval(), 3)
        # Etc etc etc

        # Test bad value
        self.assertRaises(myokit.ParseError, p, '5 beans')

    def test_parse_expression_string(self):
        # Test :meth:`parse_expression_string()`.
        from myokit._parsing import parse_expression_string
        e = parse_expression_string('5 --- 2')
        self.assertIsInstance(e, myokit.Minus)
        self.assertEqual(e.eval(), 3)

        # Only a single expression can be given
        self.assertRaisesRegex(
            myokit.ParseError, 'Unexpected token INTEGER',
            parse_expression_string, '5 + 2 3 * 7')

    def test_parse_state(self):
        # Test parse_state()
        # Further tests in test_loadsave

        # Test basic
        from myokit._parsing import parse_state
        self.assertEqual(parse_state(''), [])

        code = (
            '1.23\n',
        )
        self.assertEqual(parse_state(code), [1.23])

        code = (
            '1.23\n',
            '5.67',
        )
        self.assertEqual(parse_state(code), [1.23, 5.67])

        code = (
            'x.y = 1.23\n',
        )
        self.assertEqual(parse_state(code), {'x.y': 1.23})

        code = (
            'x.y = 1.23\n',
            'x.z = 2.34',
        )
        self.assertEqual(parse_state(code), {'x.y': 1.23, 'x.z': 2.34})

        # Must use fully qualified names
        code = (
            'x = 5',
        )
        self.assertRaisesRegex(
            myokit.ParseError, 'must be fully qualified', parse_state, code)

    def test_format_parse_error(self):
        # Test format_parse_error.

        # Test basic formatting, with and without source
        bad = '    5 + / 2'
        try:
            myokit.parse_expression(bad)
        except myokit.ParseError as e:

            # No source
            self.assertEqual(
                myokit.format_parse_error(e),
                '\n'.join([
                    'Syntax error',
                    '  Unexpected token SLASH "/" expecting expression',
                    'On line 1 character 8',
                ])
            )

            # List-of-strings source
            self.assertEqual(
                myokit.format_parse_error(e, source=[bad]),
                '\n'.join([
                    'Syntax error',
                    '  Unexpected token SLASH "/" expecting expression',
                    'On line 1 character 8',
                    '  5 + / 2',
                    '      ^'
                ])
            )

            # File source
            with TemporaryDirectory() as d:
                path = d.path('mmt')
                with open(path, 'w') as f:
                    f.write(bad + '\n')
                myokit.format_parse_error(e, source=path),
                '\n'.join([
                    'Syntax error',
                    '  Unexpected token SLASH "/" expecting expression',
                    'On line 1 character 8',
                    '  5 + / 2',
                    '      ^'
                ])

            # Line doesn't exist in source
            self.assertEqual(
                myokit.format_parse_error(e, source=[]),
                '\n'.join([
                    'Syntax error',
                    '  Unexpected token SLASH "/" expecting expression',
                    'On line 1 character 8',
                ])
            )

            # Char doesn't exist in source
            self.assertEqual(
                myokit.format_parse_error(e, source=['x']),
                '\n'.join([
                    'Syntax error',
                    '  Unexpected token SLASH "/" expecting expression',
                    'On line 1 character 8',
                ])
            )

        # Very long lines
        bad = '    1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + 100 + 1000 + 11'
        bad += ' + 12 + 13 + 14 + 15 + 16 + 17 + 18 + 19 + 20 + 21 + 22'
        bad += ' + 23 + 24 + 25 + 26 + 27 + 28 + 29 + 30 + 31'

        # Error near start
        error = '\n'.join([
            'Syntax error',
            '  Unexpected token SLASH "/" expecting expression',
            'On line 1 character 12',
            '  1 + 2 + / 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + 100 + 1000 + ..',
            '          ^',
        ])
        b = bad[:12] + '/ ' + bad[12:]
        try:
            myokit.parse_expression(b)
        except myokit.ParseError as e:
            self.assertEqual(myokit.format_parse_error(e, source=[b]), error)

        error = '\n'.join([
            'Syntax error',
            '  Unexpected token SLASH "/" expecting expression',
            'On line 1 character 83',
            '  ..+ 12 + 13 + 14 + 15 + / 16 + 17 + 18 + 19 + 20 + 21 + 22..',
            '                          ^',
        ])
        b = bad[:83] + '/ ' + bad[83:]
        try:
            myokit.parse_expression(b)
        except myokit.ParseError as e:
            self.assertEqual(myokit.format_parse_error(e, source=[b]), error)

        error = '\n'.join([
            'Syntax error',
            '  Unexpected token SLASH "/" expecting expression',
            'On line 1 character 133',
            '  ..+ 21 + 22 + 23 + 24 + 25 + / 26 + 27 + 28 + 29 + 30 + 31',
            '                               ^',
        ])
        b = bad[:133] + '/ ' + bad[133:]
        try:
            myokit.parse_expression(b)
        except myokit.ParseError as e:
            self.assertEqual(myokit.format_parse_error(e, source=[b]), error)


class ModelParseTest(unittest.TestCase):
    """
    Test parsing models.
    """

    def test_model_creation(self):
        # Basic model creation via the parser
        m = myokit.load_model(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Test components
        self.assertEqual(len(m), 9)
        self.assertIn('engine', m)
        self.assertIn('membrane', m)
        self.assertIn('cell', m)
        self.assertIn('ina', m)
        self.assertIn('ik1', m)
        self.assertIn('ica', m)
        self.assertIn('ib', m)
        self.assertIn('ik', m)
        self.assertIn('ikp', m)

        # Test state
        states = [
            'membrane.V',
            'ina.m',
            'ina.h',
            'ina.j',
            'ica.d',
            'ica.f',
            'ik.x',
            'ica.Ca_i']
        values = [
            -84.5286,
            0.0017,
            0.9832,
            0.995484,
            0.000003,
            1,
            0.0057,
            0.0002]
        self.assertEqual(len(values), m.count_states())

        # Test initial value parsing
        out = ', '.join([str(x) for x in m.states()])
        ref = ', '.join(states)
        self.assertEqual(ref, out)
        for val, ref in zip(m.initial_values(as_floats=True), values):
            self.assertEqual(val, ref)

    def test_unresolved_reference_error(self):
        # Test unresolved reference errors.
        code = """
            [[model]]

            [engine]
            time = 0 bind time

            [c]
            x = a
            """
        self.assertRaises(myokit.ParseError, myokit.parse, code)
        try:
            myokit.parse(code)
        except myokit.ParseError as e:
            self.assertIsInstance(e.cause, myokit.IntegrityError)
            self.assertIsInstance(e.cause, myokit.UnresolvedReferenceError)
            # Token is not set when resolving name (string)
            self.assertIsNone(e.cause.token())
            # But line and char should be set in parse error by parser
            self.assertEqual(e.line, 8)
            self.assertEqual(e.char, 16)
            # Check str() method (only implemented for parse errors)
            str(e)

    def test_cyclical_reference_error(self):
        # Test cyclical reference errors.

        code = """
            [[model]]

            [engine]
            time = 0 bind time

            [c]
            x = y
            y = x
            """
        self.assertRaises(myokit.ParseError, myokit.parse, code)
        with self.assertRaises(myokit.ParseError) as e:
            myokit.parse(code)
        e = e.exception
        self.assertIsInstance(e.cause, myokit.IntegrityError)
        self.assertIsInstance(e.cause, myokit.CyclicalDependencyError)
        self.assertIn(e.line, [8, 9])
        self.assertEqual(e.char, 12)
        from myokit._parsing import NAME
        if e.line == 8:
            self.assertEqual(e.cause.token(), (NAME, 'x', 8, 12))
        else:
            self.assertEqual(e.cause.token(), (NAME, 'y', 9, 12))

    def test_piecewise(self):
        # Test a model with a piecewise statement

        m = myokit.load_model(
            os.path.join(DIR_DATA, 'conditional.mmt'))
        # Test evaluation
        x = m.get('test.x')
        y = m.get('test.y')
        s = m.initial_values(as_floats=True)
        i = m.get('membrane.V').index()
        # Test x, xo, y, yo

        v = myokit.Name(m.get('membrane.V'))
        s[i] = -80
        m.set_initial_values(s)
        self.assertEqual(x.rhs().eval(), 3)
        self.assertEqual(y.rhs().eval(), 2)
        s[i] = -10
        m.set_initial_values(s)
        self.assertEqual(x.rhs().eval(), 2)
        self.assertEqual(y.rhs().eval(), 2)
        s[i] = 30
        m.set_initial_values(s)
        self.assertEqual(x.rhs().eval(), 1)
        self.assertEqual(y.rhs().eval(), 1)
        # Test code() output by cloning
        m.clone()

    def test_initial_values(self):
        # Test if expressions for initial values are handled correctly.

        code = """
            [[model]]
            c.p = 1.0
            c.q = 10 * 2

            [engine]
            time = 0 bind time

            [c]
            dot(p) = 1
            dot(q) = 2
            """
        m = myokit.parse_model(code)
        self.assertEqual(m.get('c.p').initial_value(), myokit.Number(1))
        self.assertEqual(m.get('c.q').initial_value().code(), '10 * 2')

        # initial expression
        code = """
            [[model]]
            c.q = 10 * 2 + c.p
            c.r = c.p + c.a

            [engine]
            time = 0 bind time

            [c]
            p = 0.5
            dot(q) = 2
            dot(r) = 1
            a = p + 1
            """
        m = myokit.parse_model(code)
        self.assertEqual(
            m.get('c.q').initial_value().code(), '10 * 2 + c.p')

        # initial expression with just literals
        code = """
            [[model]]
            c.q = 10 * exp(2)

            [engine]
            time = 0 bind time

            [c]
            p = 0.5
            dot(q) = 2
            """
        m = myokit.parse_model(code)
        self.assertEqual(
            m.get('c.q').initial_value().code(), '10 * exp(2)')

        # non-existent variable in initial value
        code = """
            [[model]]
            c.p = 1.0
            c.q = 10 * 2 + c.b

            [engine]
            time = 0 bind time

            [c]
            dot(p) = 1
            dot(q) = 2
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'Unresolved reference', myokit.parse, code)

        # initial expression using child variables should fail
        code = """
            [[model]]
            c.q = a

            [engine]
            time = 0 bind time

            [c]
            p = 0.5
            dot(q) = 2
                a = 10 * 2 + p
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'Unresolved reference', myokit.parse, code)

        # non-constant initial value
        code = """
            [[model]]
            c.p = engine.time

            [engine]
            time = 0 bind time

            [c]
            dot(p) = 1
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'is not constant', myokit.parse, code)

        # indirect non-constant initial value
        code = """
            [[model]]
            c.p = c.r
            c.q = 1

            [engine]
            time = 0 bind time

            [c]
            dot(p) = 1
            dot(q) = 2
            r = 2 * q
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'is not constant', myokit.parse, code)

        # looks like it shouldn't be constant, but it is!
        code = """
            [[model]]
            c.p = dot(c.p)

            [engine]
            time = 0 bind time

            [c]
            dot(p) = 0
            """
        # Treat all state variables as non-constant
        self.assertRaisesRegex(
            myokit.ParseError, 'is not constant', myokit.parse, code)

        # Initial value cycles are caught in validation
        code = """
            [[model]]
            x.x = x.x

            [x]
            dot(x) = x
            t = 0 bind time
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'is not constant', myokit.parse, code)

    def test_aliases(self):
        code = """
            [[model]]

            [engine]
            time = 0 bind time

            [c]
            p = 1

            [d]
            use c.p
            q = 2 * p

            [e]
            use c.p as ploep
            r = 10 * ploep
            """
        myokit.parse(code)

        # Duplicate alias is allowed
        code = """
            [[model]]

            [engine]
            time = 0 bind time

            [c]
            p = 1

            [d]
            use c.p as one
            use c.p as two
            q = 2 * one
            """
        myokit.parse(code)

        # Duplicate name
        code = """
            [[model]]

            [engine]
            time = 0 bind time

            [c]
            p = 1

            [d]
            q = 10
            use c.p as q
            """
        self.assertRaises(myokit.ParseError, myokit.parse, code)

        # API Alias
        m = myokit.Model('InvalidNameAlias')
        c = m.add_component('c')
        p = c.add_variable('p')
        p.set_rhs('4 / 20')
        q = c.add_variable('q')
        q.set_rhs('-sqrt(3)')
        d = m.add_component('d')
        r = d.add_variable('r')
        r.set_rhs('1 / 1e2')
        d.add_alias('p', p)

        # Invalid name
        self.assertRaises(myokit.InvalidNameError, d.add_alias, '_plf', q)

        # Duplicate names
        self.assertRaises(myokit.DuplicateName, d.add_alias, 'p', q)
        self.assertRaises(myokit.DuplicateName, c.add_alias, 'p', r)

    def test_clone_code_parse(self):
        # Test the cloning, code and parse() by exporting models and
        # reading them in again.
        models = [
            'conditional.mmt',
        ]
        for model in models:
            m1 = myokit.load_model(os.path.join(DIR_DATA, model))
            c1 = m1.code()
            m2 = myokit.parse(c1)[0]
            c2 = m2.code()
            self.assertEqual(c1, c2)
            m3 = m2.clone()
            c3 = m3.code()
            self.assertEqual(c1, c3)

    def test_advanced_units(self):
        # Test the new unit syntax where literals have units.
        model = 'beeler-1977-units.mmt'
        m = myokit.load_model(os.path.join(DIR_DATA, model))
        m.validate()

    def test_unresolved_references(self):
        # Test parsing models with unresolved references.
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = 5
            """
        myokit.parse_model(m)

        # Bad variable in same component
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = 5
            y = z
            """
        self.assertRaises(myokit.ParseError, myokit.parse_model, m)

        # Bad component
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = 5
            y = d.z
            """
        self.assertRaises(myokit.ParseError, myokit.parse_model, m)

        # Bad variable in other component
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = 5
            y = d.b

            [d]
            a = 12
            """
        self.assertRaises(myokit.ParseError, myokit.parse_model, m)

    def test_invalid_dot_in_rhs(self):
        # Test parsing a model with an invalid dot() in a variable's RHS.

        # This model has dot(1 - x), which is not allowed!
        code = """
            [[model]]
            c.x = 0

            [engine]
            time = 0 bind time

            [c]
            dot(x) = 1 - x
            p = dot(1 - x)
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'only be used on variables', myokit.parse, code)

    def test_strip_expression_units(self):
        # Test :meth:`strip_expression_units`.
        from myokit._parsing import parse_model, strip_expression_units

        m1 = myokit.load_model('example')
        m2 = parse_model(m1.code())
        self.assertEqual(m1.code(), m2.code())

        m2 = parse_model(strip_expression_units(m1.code()))
        c1 = m1.code()
        c2 = m2.code()
        self.assertNotEqual(c1, c2)
        self.assertTrue(len(c2) < len(c1))
        self.assertEqual(
            m1.evaluate_derivatives(), m2.evaluate_derivatives())

        m2 = parse_model(strip_expression_units(m1.code().splitlines()))
        c1 = m1.code()
        c2 = m2.code()
        self.assertNotEqual(c1, c2)
        self.assertTrue(len(c2) < len(c1))
        self.assertEqual(
            m1.evaluate_derivatives(), m2.evaluate_derivatives())

    def test_function_parsing(self):
        # Test parsing of functions.

        # Test simple function
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = asin(sin(1))
            """
        m = myokit.parse_model(m)
        self.assertAlmostEqual(m.get('c.x').eval(), 1)

        # Test function with multiple arguments
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = log(8, 2)
            """
        m = myokit.parse_model(m)
        self.assertAlmostEqual(m.get('c.x').eval(), 3)

        # Unknown function
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = blog(8, 2)
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'Unknown function', myokit.parse_model, m)

        # Wrong number of arguments
        m = """
            [[model]]

            [c]
            t = 0 bind time
            x = sin(8, 2)
            """
        self.assertRaisesRegex(
            myokit.ParseError, 'Wrong number of arguments', myokit.parse_model,
            m)


if __name__ == '__main__':
    unittest.main()



#
# Provides Stan support (see: http://mc-stan.org/)
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import StanExporter
from ._ewriter import StanExpressionWriter


# Importers
# Exporters
_exporters = {
    'stan': StanExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'stan': StanExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Language keywords
# Variable names can overlap with built-in function names in stan, so that
# `real abs = abs(-5)` is a perfectly valid statement.
keywords = [
    # Data types, constraints
    'cholesky_factor_corr',
    'cholesky_factor_cov',
    'corr_matrix',
    'cov_matrix',
    'int',
    'lower',
    'matrix',
    'ordered',
    'positive_ordered',
    'real',
    'row_vector',
    'simplex',
    'unit_vector',
    'upper',
    'vector',
    #not a number
    #+inf
    #-inf
    # Block identifiers
    'data',
    'functions',
    'generated_quantities',
    'model',
    'parameters',
    'transformed data',
    'transformed_parameters',
    # Language
    'break',
    'continue',
    'else',
    'false',
    'for',
    'if',
    'in',
    'repeat',
    'return',
    'then',
    'true',
    'until',
    'while',
    # Reserved names from C++ implementation
    'fvar',
    'STAN_MAJOR',
    'STAN_MATH_MAJOR',
    'STAN_MATH_MINOR',
    'STAN_MATH_PATH',
    'STAN_MINOR',
    'STAN_PATCH',
    'var',
    # Further C++ keywords
    'alignas',
    'alignof',
    'and',
    'and_eq',
    'asm',
    'auto',
    'bitand',
    'bitor',
    'bool',
    'break',
    'case',
    'catch',
    'char',
    'char16_t',
    'char32_t',
    'class',
    'compl',
    'const',
    'const_cast',
    'constexpr',
    'continue',
    'decltype',
    'default',
    'delete',
    'do',
    'double',
    'dynamic_cast',
    'else',
    'enum',
    'explicit',
    'export',
    'extern',
    'false',
    'float',
    'for',
    'friend',
    'goto',
    'if',
    'inline',
    'int',
    'long',
    'mutable',
    'namespace',
    'new',
    'noexcept',
    'not',
    'not_eq',
    'nullptr',
    'operator',
    'or',
    'or_eq',
    'private',
    'protected',
    'public',
    'register',
    'reinterpret_cast',
    'return',
    'short',
    'signed',
    'sizeof',
    'static',
    'static_assert',
    'static_cast',
    'struct',
    'switch',
    'template',
    'this',
    'thread_local',
    'throw',
    'true',
    'try',
    'typedef',
    'typeid',
    'typename',
    'union',
    'unsigned',
    'using',
    'virtual',
    'void',
    'volatile',
    'wchar_t',
    'while',
    'xor',
    'xor_eq',
]



#!/usr/bin/env python3
#
# Tests the lib.deps dependency graphing module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import random
import unittest

import myokit
import myokit.lib.deps as deps


class LibDepsTest(unittest.TestCase):

    def test_state_dependency_matrix(self):
        # Test create_ and plot_ state dependency matrix method, to create
        # matrix plot of state interdependencies.

        # Load model
        model = myokit.load_model('example')

        # Test direct version (simplest matrix)
        matrix = deps.create_state_dependency_matrix(model, direct=True)
        self.assertEqual(matrix, [
            # V m  h  j  d  f  x Cai
            [1, 1, 1, 1, 1, 1, 1, 1],  # V
            [1, 1, 0, 0, 0, 0, 0, 0],  # m
            [1, 0, 1, 0, 0, 0, 0, 0],  # h
            [1, 0, 0, 1, 0, 0, 0, 0],  # j
            [1, 0, 0, 0, 1, 0, 0, 0],  # d
            [1, 0, 0, 0, 0, 1, 0, 0],  # f
            [1, 0, 0, 0, 0, 0, 1, 0],  # x
            [1, 0, 0, 0, 1, 1, 0, 1],  # Cai depends on d and f
        ])

        # Test direct version without V
        matrix = deps.create_state_dependency_matrix(
            model, direct=True, knockout=['membrane.V'])
        self.assertEqual(matrix, [
            # V m  h  j  d  f  x Cai
            [0, 0, 0, 0, 0, 0, 0, 0],  # V
            [0, 1, 0, 0, 0, 0, 0, 0],  # m
            [0, 0, 1, 0, 0, 0, 0, 0],  # h
            [0, 0, 0, 1, 0, 0, 0, 0],  # j
            [0, 0, 0, 0, 1, 0, 0, 0],  # d
            [0, 0, 0, 0, 0, 1, 0, 0],  # f
            [0, 0, 0, 0, 0, 0, 1, 0],  # x
            [0, 0, 0, 0, 1, 1, 0, 1],  # Cai depends on d and f
        ])
        self.assertRaisesRegex(
            ValueError, 'must be states', deps.create_state_dependency_matrix,
            model, direct=True, knockout=['ina.INa'])

        # Test indirect version
        matrix = deps.create_state_dependency_matrix(model, direct=False)
        self.assertEqual(matrix, [
            # V m  h  j  d  f  x Cai
            [1, 1, 1, 1, 1, 1, 1, 1],  # V
            [1, 1, 2, 2, 2, 2, 2, 2],  # m
            [1, 2, 1, 2, 2, 2, 2, 2],  # h
            [1, 2, 2, 1, 2, 2, 2, 2],  # j
            [1, 2, 2, 2, 1, 2, 2, 2],  # d
            [1, 2, 2, 2, 2, 1, 2, 2],  # f
            [1, 2, 2, 2, 2, 2, 1, 2],  # x
            [1, 2, 2, 2, 1, 1, 2, 1],  # Cai depends on d and f
        ])

        # Test indirect version, without V
        matrix = deps.create_state_dependency_matrix(
            model, direct=False, knockout=['membrane.V'])
        self.assertEqual(matrix, [
            # V m  h  j  d  f  x Ca_i
            [0, 0, 0, 0, 0, 0, 0, 0],  # V
            [0, 1, 0, 0, 0, 0, 0, 0],  # m
            [0, 0, 1, 0, 0, 0, 0, 0],  # h
            [0, 0, 0, 1, 0, 0, 0, 0],  # j
            [0, 0, 0, 0, 1, 0, 0, 0],  # d
            [0, 0, 0, 0, 0, 1, 0, 0],  # f
            [0, 0, 0, 0, 0, 0, 1, 0],  # x
            [0, 0, 0, 0, 1, 1, 0, 1],  # Cai depends on d and f

        ])

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')

        # Create plot
        deps.plot_state_dependency_matrix(model)

    def test_component_dependency_graph(self):
        # Test create_ and plot_ component dependency graph method, to show
        # digraph plot of variable dependencies.

        # Nodes are placed at random, so seed the generator for consistent
        # output (otherwise we risk missing full cover on some runs).
        random.seed(1)

        # Load model
        model = myokit.load_model('example')

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')

        # Create plot
        self.assertFalse(model.has_interdependent_components())
        deps.plot_component_dependency_graph(model)

        # Add cyclical component dependencies, plot again
        # This time, the acylical plotting code can't be used.
        v = model.get('ina').add_variable('ion')
        v.set_rhs('membrane.i_ion')
        self.assertTrue(model.has_interdependent_components())
        deps.plot_component_dependency_graph(model)

    def test_variable_dependency_graph(self):
        # Test create_ and plot_ variable dependency graph method, to show
        # digraph plot of variable dependencies.

        # Nodes are placed at random, so seed the generator for consistent
        # output (otherwise we risk missing full cover on some runs).
        random.seed(1)

        # Load model
        model = myokit.load_model('example')

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')

        # Create plot
        deps.plot_variable_dependency_graph(model)


class DiGraphTest(unittest.TestCase):
    """
    Tests parts of :class:`myokit.lib.deps.DiGraph`.
    """

    def test_basic(self):
        # Test basic DiGraph functions.

        # Create empty graph
        d = deps.DiGraph()
        self.assertEqual(len(d), 0)
        self.assertEqual(d.text(), 'Empty graph')
        d.add_node(1)
        self.assertEqual(len(d), 1)
        d.add_node(2)
        self.assertEqual(len(d), 2)
        d.add_node(3)
        self.assertEqual(len(d), 3)
        d.add_edge(1, 2)
        self.assertEqual(len(d), 3)
        d.add_edge(2, 3)
        d.add_edge(3, 1)

        # Node errors
        self.assertRaisesRegex(ValueError, 'Duplicate', d.add_node, 3)
        dwrong = deps.DiGraph()
        xwrong = dwrong.add_node(99)
        self.assertRaisesRegex(ValueError, 'another graph', d.add_node, xwrong)
        self.assertRaisesRegex(
            ValueError, 'another graph', d.uid_or_node, xwrong)

        # Edge errors
        self.assertRaisesRegex(ValueError, 'Node not found', d.add_edge, 1, 4)

        # Test text
        self.assertEqual(
            d.text(), 'Node "1"\n  > Node "2"\nNode "2"\n  > Node "3"\n'
            'Node "3"\n  > Node "1"')
        self.assertEqual(
            d.text(True), '0, 1, 0\n0, 0, 1\n1, 0, 0')

        # Test cloning
        d2 = deps.DiGraph(d)
        self.assertEqual(len(d2), 3)
        self.assertEqual(d.text(), d2.text())

        # Test build from matrix: Doesn't copy labels
        matrix = d.matrix()
        d3 = deps.DiGraph(matrix)
        self.assertEqual(len(d3), 3)
        self.assertEqual(
            d3.text(), 'Node "0"\n  > Node "1"\nNode "1"\n  > Node "2"\n'
            'Node "2"\n  > Node "0"')
        d3.build_from_matrix(matrix)
        self.assertEqual(
            d3.text(), 'Node "0"\n  > Node "1"\nNode "1"\n  > Node "2"\n'
            'Node "2"\n  > Node "0"')
        d3.build_from_matrix(matrix, edges_only=True)
        self.assertRaises(ValueError, deps.DiGraph, matrix[:-1])
        self.assertEqual(
            d3.text(), 'Node "0"\n  > Node "1"\nNode "1"\n  > Node "2"\n'
            'Node "2"\n  > Node "0"')
        self.assertRaises(ValueError, d.build_from_matrix, matrix[:-1])
        matrix = [x[:-1] for x in matrix[:-1]]
        self.assertRaises(
            ValueError, d.build_from_matrix, matrix, edges_only=True)

        # Test that megs and layering fail on cyclical graph
        self.assertRaisesRegex(Exception, 'cyclical', d.cg_layers_dag)
        d.add_edge(1, 1)
        self.assertRaisesRegex(Exception, 'self-referencing', d.meg_dag)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Test module
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
# The TemporaryDirectory class was adapted from Pints
# See: https://github.com/pints-team/pints
#
import os
import tempfile
import warnings

import myokit


# The test directory
DIR_TEST = os.path.abspath(os.path.dirname(__file__))

# The data directory
DIR_DATA = os.path.join(DIR_TEST, 'data')

# Extra files in the data directory for load/save testing
DIR_IO = os.path.join(DIR_DATA, 'io')

# Extra files in the data directory for format testing
DIR_FORMATS = os.path.join(DIR_DATA, 'formats')

# OpenCL support
OpenCL_FOUND = myokit.OpenCL.available()
OpenCL_DOUBLE_PRECISION = False
OpenCL_DOUBLE_PRECISION_CONNECTIONS = False
if OpenCL_FOUND:
    info = myokit.OpenCL.current_info()
    OpenCL_DOUBLE_PRECISION = info.has_extension('cl_khr_fp64')
    if OpenCL_DOUBLE_PRECISION:
        OpenCL_DOUBLE_PRECISION_CONNECTIONS = info.has_extension(
            'cl_khr_int64_base_atomics')
    del info


class TemporaryDirectory:
    """
    ContextManager that provides a temporary directory to create temporary
    files in. Deletes the directory and its contents when the context is
    exited.
    """
    def __init__(self):
        super().__init__()
        self._dir = None

    def __enter__(self):
        self._dir = os.path.realpath(tempfile.mkdtemp())
        return self

    def path(self, path=None):
        """
        Returns an absolute path to a file or directory name inside this
        temporary directory, that can be used to write to.

        Example::

            with pints.io.TemporaryDirectory() as d:
                filename = d.path('test.txt')
                with open(filename, 'w') as f:
                    f.write('Hello')
                with open(filename, 'r') as f:
                    print(f.read())
        """
        if self._dir is None:
            raise RuntimeError(
                'TemporaryDirectory.path() can only be called from inside the'
                ' context.')

        if path is None:
            return self._dir

        path = os.path.realpath(os.path.join(self._dir, path))
        if path[0:len(self._dir)] != self._dir:
            raise ValueError(
                'Relative path specified to location outside of temporary'
                ' directory: ' + path)

        return path

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            myokit.tools.rmtree(self._dir)
        finally:
            self._dir = None

    def __str__(self):
        if self._dir is None:
            return '<TemporaryDirectory, outside of context>'
        else:
            return self._dir


class TestReporter(myokit.ProgressReporter):
    """
    Progress reporter just for debugging.
    """
    def __init__(self):
        self.entered = False
        self.exited = False
        self.updated = False

    def enter(self, msg=None):
        self.entered = True

    def exit(self):
        self.exited = True

    def update(self, f):
        self.updated = True
        return True


class CancellingReporter(myokit.ProgressReporter):
    """
    Progress reporter just for debugging, dies after `x` updates.
    """
    def __init__(self, okays=0):
        self.okays = int(okays)

    def enter(self, msg=None):
        pass

    def exit(self):
        pass

    def update(self, f):
        self.okays -= 1
        return self.okays >= 0


class WarningCollector:
    """
    Wrapper around warnings.catch_warnings() that gathers all messages into a
    single string.
    """
    def __init__(self):
        self._warnings = []
        self._w = warnings.catch_warnings(record=True)

    def __enter__(self):
        self._warnings = self._w.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        self._w.__exit__(type, value, traceback)

    def count(self):
        """Returns the number of warnings caught."""
        return len(self._warnings)

    def has_warnings(self):
        """Returns ``True`` if there were any warnings."""
        return len(self._warnings) > 0

    def text(self):
        """Returns the text of all gathered warnings."""
        return ' '.join(str(w.message) for w in self._warnings)

    def warnings(self):
        """Returns all gathered warning objects."""
        return self._warnings


def test_case_pk_model(parameters, times):
    """
    This function returns the analytic solution of the state dynamics and
    partial derivatives for a repeated bolus infusion into a compartment with
    linear clearance.

    For details of the derivation please see the technical notes.

    Parameters
    ----------
    parameters
        A list with the initial drug amount, the elimination rate and the
        dose rate and the infusion duration at each injection.
    times
        The times for evaluation.
    """
    import numpy as np

    times = np.asarray(times)

    # Unpack parameters
    a_0, elimination_rate, dose_rate, duration = parameters

    # Compute time since start last infusion
    delta_times = times - np.floor(times)

    # Create a mask for unfinished doses
    mask = delta_times % 1 < duration

    # Compute times since stop last infusion
    delta_times -= duration

    # Compute a max and da_max / delimination_rate
    a_max = \
        dose_rate * (1 - np.exp(-elimination_rate * duration)) \
        / elimination_rate / (1 - np.exp(-elimination_rate))
    da_max = \
        - a_max * (
            1 / elimination_rate
            + np.exp(-elimination_rate) / (
                1 - np.exp(-elimination_rate))) \
        + duration * dose_rate * (
            np.exp(-elimination_rate * duration)
            / (1 - np.exp(-elimination_rate))
            / elimination_rate)

    # Compute amount
    amount = a_0 * np.exp(-elimination_rate * times)
    amount[mask] += \
        a_max * (
            np.exp(-elimination_rate * (delta_times[mask] + 1))
            - np.exp(-elimination_rate * (times[mask] - duration + 1))) \
        + dose_rate / elimination_rate * (
            1 - np.exp(-elimination_rate * (delta_times[mask] + duration)))
    amount[~mask] += \
        a_max * (
            np.exp(-elimination_rate * delta_times[~mask])
            - np.exp(-elimination_rate * (times[~mask] - duration + 1)))

    # Compute partials
    damount_dinitial_amount = np.exp(-elimination_rate * times)
    damount_delimination_rate = \
        -times * a_0 * np.exp(-elimination_rate * times)
    damount_delimination_rate[mask] += \
        - a_max * (
            (delta_times[mask] + 1)
            * np.exp(-elimination_rate * (delta_times[mask] + 1))
            - (times[mask] - duration + 1)
            * np.exp(-elimination_rate * (times[mask] - duration + 1))) \
        + da_max * (
            np.exp(-elimination_rate * (delta_times[mask] + 1))
            - np.exp(-elimination_rate * (times[mask] - duration + 1))) \
        + dose_rate / elimination_rate * (
            delta_times[mask] + duration + 1 / elimination_rate) \
        * np.exp(-elimination_rate * (delta_times[mask] + duration)) \
        - dose_rate / elimination_rate**2
    damount_delimination_rate[~mask] += \
        - a_max * (
            delta_times[~mask]
            * np.exp(-elimination_rate * delta_times[~mask])
            - (times[~mask] - duration + 1) *
            np.exp(-elimination_rate * (times[~mask] - duration + 1))) \
        + da_max * (
            np.exp(-elimination_rate * delta_times[~mask])
            - np.exp(-elimination_rate * (times[~mask] - duration + 1)))
    partials = np.vstack([damount_dinitial_amount, damount_delimination_rate])

    return amount, partials



#!/usr/bin/env python3
#
# Tests the exporters from the format module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.formats

from myokit.tests import TemporaryDirectory


class ExportTest(unittest.TestCase):
    """
    Tests various exporters.
    """

    def _test(self, e, model=None, protocol=None):
        """ Test a given exporter `e`. """

        # Test info method.
        self.assertIsInstance(e.post_export_info(), str)

        # Load model, protocol
        m, p = model, protocol
        if m is None:
            m = myokit.load_model('example')
        if p is None:
            p = myokit.load_protocol('example')

        with TemporaryDirectory() as d:
            path = d.path()

            # Try exports
            exports = 0

            # Try model export
            if e.supports_model():
                exports += 1

                # Basic export
                fpath = os.path.join(path, 'model.txt')
                ret = e.model(fpath, m)
                self.assertIsNone(ret)
                self.assertTrue(os.path.isfile(fpath))

                # Unnamed model
                name = m.name()
                try:
                    m.set_name(None)
                    ret = e.model(fpath, m)
                    self.assertIsNone(ret)
                    self.assertTrue(os.path.isfile(fpath))
                finally:
                    m.set_name(name)

            else:
                self.assertRaises(NotImplementedError, e.model, path, m)

            # Try runnable export
            if e.supports_runnable():
                exports += 1

                # Check without protocol
                dpath = os.path.join(path, 'runnable1')
                ret = e.runnable(dpath, m)
                self.assertIsNone(ret)
                self.assertTrue(os.path.isdir(dpath))
                self.assertTrue(len(os.listdir(dpath)) > 0)

                # Check with protocol
                dpath = os.path.join(path, 'runnable2')
                ret = e.runnable(dpath, m, p)
                self.assertIsNone(ret)
                self.assertTrue(os.path.isdir(dpath))
                self.assertTrue(len(os.listdir(dpath)) > 0)

                # Write to complex path
                dpath = os.path.join(path, 'runnable3', 'nest', 'test')
                ret = e.runnable(dpath, m, p)
                self.assertIsNone(ret)
                self.assertTrue(os.path.isdir(dpath))
                self.assertTrue(len(os.listdir(dpath)) > 0)

            else:

                self.assertRaises(NotImplementedError, e.runnable, path, m, p)

            # Test if any exports were available
            if exports == 0:
                raise Exception(
                    'No types of export supported by: ' + exporter)

    def test_runnable_exporter_shared(self):
        # Test shared functionality of the TemplatedRunnableExporters.

        e = myokit.formats.exporter('ansic')

        # Load model, protocol
        m, p, x = myokit.load('example')

        # Create empty output directory as subdirectory of DIR_OUT
        with TemporaryDirectory() as d:
            path = d.path()

            # Simple export
            dpath = os.path.join(path, 'runnable1')
            ret = e.runnable(dpath, m)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # Write to complex path
            dpath = os.path.join(path, 'runnable2', 'nest', 'test')
            ret = e.runnable(dpath, m, p)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # Overwrite existing path
            ret = e.runnable(dpath, m, p)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # Path pointing to file
            dpath = os.path.join(path, 'file')
            with open(dpath, 'w') as f:
                f.write('contents\n')
            self.assertRaisesRegex(
                myokit.ExportError, 'file exists', e.runnable, dpath, m, p)

            # Directory exists where we're trying to write a file
            dpath = os.path.join(path, 'runnable3')
            fname = os.path.join(dpath, 'sim.c')
            os.makedirs(fname)
            self.assertRaisesRegex(
                myokit.ExportError, 'Directory exists',
                e.runnable, dpath, m, p)

            # Directory embedded in the output file path
            def embedded():
                return {'sim.c': 'nested/sim.c'}

            # 1. Normal operation
            e._dict = embedded
            dpath = os.path.join(path, 'runnable4')
            ret = e.runnable(dpath, m, p)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # 2. Try to create directory where file exists
            def embedded():
                return {'sim.c': 'nested/sim.c/som.c'}

            e._dict = embedded
            dpath = os.path.join(path, 'runnable4')
            self.assertRaisesRegex(
                myokit.ExportError, 'file or link', e.runnable, dpath, m, p)

    def test_completeness(self):
        # Test that all exporters have a test (so meta!).

        methods = [x for x in dir(self) if x[:5] == 'test_']
        for name in myokit.formats.exporters():
            name = name.replace('-', '_')
            name = 'test_' + name + '_exporter'
            self.assertIn(name, methods)

    def test_ansic_exporter(self):
        self._test(myokit.formats.exporter('ansic'))

    def test_ansic_cable_exporter(self):
        self._test(myokit.formats.exporter('ansic-cable'))

    def test_ansic_euler_exporter(self):
        self._test(myokit.formats.exporter('ansic-euler'))

    def test_cellml_exporter(self):
        self._test(myokit.formats.exporter('cellml'))

    def test_cellml1_exporter(self):
        self._test(myokit.formats.exporter('cellml1'))

    def test_cellml2_exporter(self):
        self._test(myokit.formats.exporter('cellml2'))

    def test_cuda_kernel_exporter(self):
        self._test(myokit.formats.exporter('cuda-kernel'))

    def test_cuda_kernel_rl_exporter(self):
        self._test(myokit.formats.exporter('cuda-kernel-rl'))

    def test_easyml_exporter(self):
        self._test(myokit.formats.exporter('easyml'))

    def test_html_exporter(self):
        self._test(myokit.formats.exporter('html'))

    def test_latex_article_exporter(self):
        self._test(myokit.formats.exporter('latex-article'))

    def test_latex_poster_exporter(self):
        self._test(myokit.formats.exporter('latex-poster'))

    def test_matlab_exporter(self):
        self._test(myokit.formats.exporter('matlab'))

    def test_opencl_exporter(self):
        self._test(myokit.formats.exporter('opencl'))

    def test_opencl_rl_exporter(self):
        self._test(myokit.formats.exporter('opencl-rl'))

    def test_opencl_exporter_errors(self):
        # Checks the errors raised by the OpenCL exporter.
        e = myokit.formats.exporter('opencl')

        # Label membrane_potential must be set
        m = myokit.load_model('example')
        m.label('membrane_potential').set_label(None)
        self.assertRaisesRegex(
            ValueError, 'membrane_potential', self._test, e, m)

        # Binding diffusion_current must be set
        m = myokit.load_model('example')
        m.binding('diffusion_current').set_binding(None)
        self.assertRaisesRegex(
            ValueError, 'diffusion_current', self._test, e, m)

    def test_python_exporter(self):
        self._test(myokit.formats.exporter('python'))

    def test_stan_exporter(self):
        # Basic test
        self._test(myokit.formats.exporter('stan'))

        # Test with parameters and output variable specified

        # Load model
        m = myokit.load_model('example')

        # Guess parameters
        parameters = []
        for v in m.get('ina').variables(const=True):
            if v.name()[:1] == 'p':
                parameters.append(v)
        parameters.sort(key=lambda v: myokit.tools.natural_sort_key(v.name()))

        # Set output
        output = 'ina.INa'

        # Export to stan
        e = myokit.formats.stan.StanExporter()
        with TemporaryDirectory() as d:
            dpath = d.path('out')
            ret = e.runnable(dpath, m, parameters=parameters, output=output)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

    def test_unknown_exporter(self):
        # Test the error handling for requesting an unknown exporter.
        self.assertRaisesRegex(
            KeyError, 'Exporter not found', myokit.formats.exporter, 'elvish')

    def test_xml_exporter(self):
        self._test(myokit.formats.exporter('xml'))


if __name__ == '__main__':
    unittest.main()
